"""Advanced mode command module."""

from ..events import AdvancedModeEvent
from .common import SetEnableCommand, _GetEnableCommand


class GetAdvancedMode(_GetEnableCommand):
    """Get advanced mode command."""

    name = "getAdvancedMode"
    event_type = AdvancedModeEvent


class SetAdvancedMode(SetEnableCommand):
    """Set advanced mode command."""

    name = "setAdvancedMode"
    get_command = GetAdvancedMode



"""Maps commands."""
from typing import Any, Optional, Union

from ..command import Command
from ..events import (
    MajorMapEvent,
    MapSetEvent,
    MapSetType,
    MapSubsetEvent,
    MapTraceEvent,
    MinorMapEvent,
)
from ..events.event_bus import EventBus
from ..message import HandlingResult, HandlingState
from . import CommandWithHandling
from .common import CommandResult


class GetCachedMapInfo(CommandWithHandling):
    """Get cached map info command."""

    name = "getCachedMapInfo"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        for map_status in data["info"]:
            if map_status["using"] == 1:
                return HandlingResult(
                    HandlingState.SUCCESS, {"map_id": map_status["mid"]}
                )

        return HandlingResult.analyse()

    def handle_requested(
        self, event_bus: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        result = super().handle_requested(event_bus, response)
        if result.state == HandlingState.SUCCESS and result.args:
            return CommandResult(
                result.state,
                result.args,
                [GetMapSet(result.args["map_id"], entry) for entry in MapSetType],
            )

        return result


class GetMajorMap(CommandWithHandling):
    """Get major map command."""

    name = "getMajorMap"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        values = data["value"].split(",")
        map_id = data["mid"]

        return HandlingResult(
            HandlingState.SUCCESS,
            {"map_id": map_id, "values": values},
        )

    def handle_requested(
        self, event_bus: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        result = super().handle_requested(event_bus, response)
        if result.state == HandlingState.SUCCESS and result.args:
            event_bus.notify(MajorMapEvent(True, **result.args))
            return CommandResult.success()

        return result


class GetMapSet(CommandWithHandling):
    """Get map set command."""

    _ARGS_ID = "id"
    _ARGS_SET_ID = "set_id"
    _ARGS_TYPE = "type"
    _ARGS_SUBSETS = "subsets"

    name = "getMapSet"

    def __init__(
        self,
        mid: str,
        type: Union[  # pylint: disable=redefined-builtin
            MapSetType, str
        ] = MapSetType.ROOMS,
    ) -> None:
        if isinstance(type, MapSetType):
            type = type.value

        super().__init__({"mid": mid, "type": type})

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        subsets = [subset["mssid"] for subset in data["subsets"]]
        args = {
            cls._ARGS_ID: data["mid"],
            cls._ARGS_SET_ID: data.get("msid", None),
            cls._ARGS_TYPE: data["type"],
            cls._ARGS_SUBSETS: subsets,
        }

        event_bus.notify(MapSetEvent(MapSetType(data["type"]), subsets))
        return HandlingResult(HandlingState.SUCCESS, args)

    def handle_requested(
        self, event_bus: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        result = super().handle_requested(event_bus, response)
        if result.state == HandlingState.SUCCESS and result.args:
            commands: list[Command] = []
            for subset in result.args[self._ARGS_SUBSETS]:
                commands.append(
                    GetMapSubSet(
                        mid=result.args[self._ARGS_ID],
                        msid=result.args[self._ARGS_SET_ID],
                        type=result.args[self._ARGS_TYPE],
                        mssid=subset,
                    )
                )
            return CommandResult(result.state, result.args, commands)

        return result


class GetMapSubSet(CommandWithHandling):
    """Get map subset command."""

    _ROOM_NUM_TO_NAME = {
        "0": "Default",
        "1": "Living Room",
        "2": "Dining Room",
        "3": "Bedroom",
        "4": "Study",
        "5": "Kitchen",
        "6": "Bathroom",
        "7": "Laundry",
        "8": "Lounge",
        "9": "Storeroom",
        "10": "Kids room",
        "11": "Sunroom",
        "12": "Corridor",
        "13": "Balcony",
        "14": "Gym",
    }

    name = "getMapSubSet"

    def __init__(
        self,
        *,
        mid: Union[str, int],
        mssid: Union[str, int],
        msid: Optional[Union[str, int]] = None,
        type: Union[  # pylint: disable=redefined-builtin
            MapSetType, str
        ] = MapSetType.ROOMS
    ) -> None:
        if isinstance(type, MapSetType):
            type = type.value

        if msid is None and type == MapSetType.ROOMS.value:
            raise ValueError("msid is required when type='vw'")

        super().__init__(
            {
                "mid": str(mid),
                "msid": str(msid),
                "type": type,
                "mssid": str(mssid),
            },
        )

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        if MapSetType.has_value(data["type"]):
            subtype = data.get("subtype", data.get("subType", None))

            event_bus.notify(
                MapSubsetEvent(
                    id=int(data["mssid"]),
                    type=MapSetType(data["type"]),
                    coordinates=data["value"],
                    subtype=cls._ROOM_NUM_TO_NAME[subtype] if subtype else None,
                )
            )

            return HandlingResult.success()

        return HandlingResult.analyse()


class GetMapTrace(CommandWithHandling):
    """Get map trace command."""

    _TRACE_POINT_COUNT = 200

    name = "getMapTrace"

    def __init__(self, trace_start: int = 0) -> None:
        super().__init__(
            {"pointCount": self._TRACE_POINT_COUNT, "traceStart": trace_start},
        )

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        total = int(data["totalCount"])
        start = int(data["traceStart"])

        if "traceValue" not in data:
            # todo verify that this is legit pylint: disable=fixme
            return HandlingResult.analyse()

        event_bus.notify(
            MapTraceEvent(start=start, total=total, data=data["traceValue"])
        )
        return HandlingResult(HandlingState.SUCCESS, {"start": start, "total": total})

    def handle_requested(
        self, event_bus: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        result = super().handle_requested(event_bus, response)
        if result.state == HandlingState.SUCCESS and result.args:
            start = result.args["start"] + self._TRACE_POINT_COUNT
            if start < result.args["total"]:
                return CommandResult(result.state, result.args, [GetMapTrace(start)])

        return result


class GetMinorMap(CommandWithHandling):
    """Get minor map command."""

    name = "getMinorMap"

    def __init__(self, *, map_id: str, piece_index: int) -> None:
        super().__init__({"mid": map_id, "type": "ol", "pieceIndex": piece_index})

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        if data.get("type", "ol") == "ol":
            # onMinorMap sends no type, so fallback to "ol"
            event_bus.notify(MinorMapEvent(data["pieceIndex"], data["pieceValue"]))
            return HandlingResult.success()

        return HandlingResult.analyse()



"""Battery messages."""
from typing import Any

from ..events import BatteryEvent
from ..events.event_bus import EventBus
from ..message import HandlingResult, Message


class OnBattery(Message):
    """On battery message."""

    name = "onBattery"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        event_bus.notify(BatteryEvent(data["value"]))
        return HandlingResult.success()



import inspect

import deebot_client.events
from deebot_client.events import EnableEvent, Event
from deebot_client.events.const import EVENT_DTO_REFRESH_COMMANDS


def test_events_has_refresh_function():
    for name, obj in inspect.getmembers(deebot_client.events, inspect.isclass):
        if issubclass(obj, Event) and obj not in [Event, EnableEvent]:
            assert obj in EVENT_DTO_REFRESH_COMMANDS



"""Logging filter module."""

import copy
from logging import Filter, Logger, LogRecord, getLogger
from typing import Any


class SanitizeFilter(Filter):
    """Filter to sensitive data."""

    # all lowercase
    _SANITIZE_LOG_KEYS: set[str] = {
        "auth",
        "did",
        "email",
        "login",
        "mobile",
        "token",
        "uid",
        "user",
    }

    def filter(self, record: LogRecord) -> bool:
        """Filter log record."""
        # The call signature matches string interpolation: args can be a tuple or a dict
        if isinstance(record.args, dict):
            record.args = self._sanitize_data(record.args)
        elif isinstance(record.args, tuple):
            record.args = tuple(self._sanitize_data(value) for value in record.args)

        return True

    def _sanitize_data(self, data: Any) -> Any:
        """Sanitize data (remove personal data)."""
        if isinstance(data, (set, list)):
            return [self._sanitize_data(entry) for entry in data]

        if not isinstance(data, dict):
            return data

        sanitized_data = None
        for key, value in data.items():
            if any(substring in key.lower() for substring in self._SANITIZE_LOG_KEYS):
                if sanitized_data is None:
                    sanitized_data = copy.deepcopy(data)
                sanitized_data[key] = "[REMOVED]"
            elif isinstance(value, (set, list, dict)):
                if sanitized_data is None:
                    sanitized_data = copy.deepcopy(data)
                sanitized_data[key] = self._sanitize_data(value)

        return sanitized_data if sanitized_data else data


def get_logger(name: str) -> Logger:
    """Get logger with filter."""
    logger = getLogger(name)
    logger.addFilter(SanitizeFilter())
    return logger



"""Volume command module."""

from typing import Any, Mapping

from ..events import VolumeEvent
from ..message import HandlingResult
from .common import EventBus, SetCommand, _NoArgsCommand


class GetVolume(_NoArgsCommand):
    """Get volume command."""

    name = "getVolume"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """

        event_bus.notify(
            VolumeEvent(volume=data["volume"], maximum=data.get("total", None))
        )
        return HandlingResult.success()


class SetVolume(SetCommand):
    """Set volume command."""

    name = "setVolume"
    get_command = GetVolume

    def __init__(self, volume: int, **kwargs: Mapping[str, Any]) -> None:
        # removing "total" as we don't can set it (App includes it)
        kwargs.pop("total", None)
        super().__init__({"volume": volume}, **kwargs)



"""MQTT module."""
import json
import ssl
from typing import MutableMapping, Optional

from cachetools import TTLCache
from gmqtt import Client, Subscription
from gmqtt.mqtt.constants import MQTTv311

from .authentication import Authenticator
from .commands import COMMANDS_WITH_MQTT_P2P_HANDLING, CommandWithMqttP2PHandling
from .exceptions import NotInitializedError
from .logging_filter import get_logger
from .models import Configuration, Credentials, DeviceInfo
from .vacuum_bot import VacuumBot

_LOGGER = get_logger(__name__)


def _get_subscriptions(device_info: DeviceInfo) -> list[Subscription]:
    return [
        # iot/atr/[command]]/[did]]/[class]]/[resource]/j
        Subscription(
            f"iot/atr/+/{device_info.did}/{device_info.get_class}/{device_info.resource}/j"
        ),
        # iot/p2p/[command]]/[sender did]/[sender class]]/[sender resource]
        # /[receiver did]/[receiver class]]/[receiver resource]/[q|p/[request id/j
        # [q|p] q-> request p-> response
        Subscription(
            f"iot/p2p/+/+/+/+/{device_info.did}/{device_info.get_class}/{device_info.resource}/q/+/j"
        ),
        Subscription(
            f"iot/p2p/+/{device_info.did}/{device_info.get_class}/{device_info.resource}/+/+/+/p/+/j"
        ),
    ]


class MqttClient:
    """MQTT client."""

    def __init__(self, config: Configuration, authenticator: Authenticator):
        self._config = config
        self._authenticator = authenticator
        self._subscribers: MutableMapping[str, VacuumBot] = {}
        self._port = 443
        self._hostname = f"mq-{config.continent}.ecouser.net"
        if config.country.lower() == "cn":
            self._hostname = "mq.ecouser.net"

        self._client: Optional[Client] = None
        self._received_p2p_commands: MutableMapping[
            str, CommandWithMqttP2PHandling
        ] = TTLCache(maxsize=60 * 60, ttl=60)

        # pylint: disable=unused-argument
        async def _on_message(
            client: Client, topic: str, payload: bytes, qos: int, properties: dict
        ) -> None:
            _LOGGER.debug("Got message: topic=%s; payload=%s;", topic, payload.decode())
            topic_split = topic.split("/")
            if topic.startswith("iot/atr"):
                await self._handle_atr(topic_split, payload)
            elif topic.startswith("iot/p2p"):
                self._handle_p2p(topic_split, payload)
            else:
                _LOGGER.debug("Got unsupported topic: %s", topic)

        self.__on_message = _on_message

        def on_credentials_changed(credentials: Credentials) -> None:
            if self._client:
                self._client.set_auth_credentials(
                    credentials.user_id, credentials.token
                )

        authenticator.subscribe(on_credentials_changed)

    async def initialize(self) -> None:
        """Initialize MQTT."""
        if self._client:
            await self.disconnect()

        credentials = await self._authenticator.authenticate()
        client_id = f"{credentials.user_id}@ecouser/{self._config.device_id}"
        self._client = Client(client_id)
        self._client.on_message = self.__on_message
        self._client.set_auth_credentials(credentials.user_id, credentials.token)

        ssl_ctx = ssl.create_default_context()
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE
        await self._client.connect(
            self._hostname, self._port, ssl=ssl_ctx, version=MQTTv311
        )

    async def subscribe(self, vacuum_bot: VacuumBot) -> None:
        """Subscribe for messages for given vacuum."""
        if self._client is None:
            raise NotInitializedError

        device_info = vacuum_bot.device_info
        self._client.subscribe(_get_subscriptions(device_info))
        self._subscribers[device_info.did] = vacuum_bot

    def unsubscribe(self, vacuum_bot: VacuumBot) -> None:
        """Unsubscribe given vacuum."""
        device_info = vacuum_bot.device_info

        if self._subscribers.pop(device_info.did, None) and self._client:
            for subscription in _get_subscriptions(device_info):
                self._client.unsubscribe(subscription.topic)

    async def disconnect(self) -> None:
        """Disconnect from MQTT."""
        if self._client:
            await self._client.disconnect()
        self._subscribers.clear()

    async def _handle_atr(self, topic_split: list[str], payload: bytes) -> None:
        try:
            bot = self._subscribers.get(topic_split[3])
            if bot:
                data = json.loads(payload)
                await bot.handle_message(topic_split[2], data)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.error(
                "An exception occurred during handling atr message", exc_info=True
            )

    def _handle_p2p(self, topic_split: list[str], payload: bytes) -> None:
        try:
            command_name = topic_split[2]
            command_type = COMMANDS_WITH_MQTT_P2P_HANDLING.get(command_name, None)
            if command_type is None:
                # command does not support p2p handling (yet)
                return

            is_request = topic_split[9] == "q"
            request_id = topic_split[10]

            if is_request:
                payload_json = json.loads(payload)
                try:
                    data = payload_json["body"]["data"]
                except KeyError:
                    _LOGGER.warning(
                        "Could not parse p2p payload: topic=%s; payload=%s",
                        "/".join(topic_split),
                        payload_json,
                    )
                    return

                self._received_p2p_commands[request_id] = command_type(**data)
            else:
                command = self._received_p2p_commands.get(request_id, None)
                if not command:
                    _LOGGER.debug(
                        "Response to command came in probably to late. requestId=%s, commandName=%s",
                        request_id,
                        command_name,
                    )
                    return

                bot = self._subscribers.get(topic_split[3])
                if bot:
                    data = json.loads(payload)
                    command.handle_mqtt_p2p(bot.events, data)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.error(
                "An exception occurred during handling p2p message", exc_info=True
            )



"""clean log commands."""
from typing import Any, Optional

from ..events import CleanJobStatus, CleanLogEntry, CleanLogEvent
from ..exceptions import DeebotError
from ..message import HandlingResult
from .common import CommandResult, CommandWithHandling, EventBus


class GetCleanLogs(CommandWithHandling):
    """Get clean logs command."""

    name = "GetCleanLogs"

    def __init__(self, count: int = 0) -> None:
        super().__init__({"count": count})

    def handle_requested(
        self, event_bus: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        if response["ret"] == "ok":
            resp_logs: Optional[list[dict]] = response.get("logs")

            # Ecovacs API is changing their API, this request may not working properly
            if resp_logs is not None and len(resp_logs) >= 0:
                logs: list[CleanLogEntry] = []
                for log in resp_logs:
                    logs.append(
                        CleanLogEntry(
                            timestamp=log["ts"],
                            image_url=log["imageUrl"],
                            type=log["type"],
                            area=log["area"],
                            stop_reason=CleanJobStatus(int(log["stopReason"])),
                            duration=log["last"],
                        )
                    )

                event_bus.notify(CleanLogEvent(logs))
                return CommandResult.success()

        return CommandResult.analyse()

    @classmethod
    def _handle_body(cls, event_bus: EventBus, body: dict[str, Any]) -> HandlingResult:
        raise DeebotError("Should never be called!")



"""Vacuum bot module."""
import asyncio
import inspect
import re
from typing import Any, Final, Optional, Union

from .api_client import ApiClient
from .command import Command
from .commands import COMMANDS_WITH_HANDLING, Clean, CommandWithHandling
from .commands.clean import CleanAction
from .commands.custom import CustomCommand
from .events import (
    CleanLogEvent,
    LifeSpanEvent,
    PositionsEvent,
    PositionType,
    StatsEvent,
    StatusEvent,
    TotalStatsEvent,
)
from .events.event_bus import EventBus
from .logging_filter import get_logger
from .map import Map
from .message import HandlingState
from .messages import MESSAGES
from .models import DeviceInfo, VacuumState

_LOGGER = get_logger(__name__)

_COMMAND_REPLACE_PATTERN = "^((on)|(off)|(report))"
_COMMAND_REPLACE_REPLACEMENT = "get"


class VacuumBot:
    """Vacuum bot representation."""

    def __init__(
        self,
        device_info: DeviceInfo,
        api_client: ApiClient,
    ):
        self.device_info: Final[DeviceInfo] = device_info
        self._api_client = api_client

        self._semaphore = asyncio.Semaphore(3)
        self._status: StatusEvent = StatusEvent(device_info.status == 1, None)

        self.fw_version: Optional[str] = None
        self.events: Final[EventBus] = EventBus(self.execute_command)

        self.map: Final[Map] = Map(self.execute_command, self.events)

        async def on_pos(event: PositionsEvent) -> None:
            if self._status == StatusEvent(True, VacuumState.DOCKED):
                return

            deebot = next(p for p in event.positions if p.type == PositionType.DEEBOT)

            if deebot:
                on_charger = filter(
                    lambda p: p.type == PositionType.CHARGER
                    and p.x == deebot.x
                    and p.y == deebot.y,
                    event.positions,
                )
                if on_charger:
                    # deebot on charger so the status should be docked... Checking
                    self.events.request_refresh(StatusEvent)

        self.events.subscribe(PositionsEvent, on_pos)

        async def on_status(event: StatusEvent) -> None:
            last_status = self._status
            self._status = event
            if (not last_status.available) and event.available:
                # bot was unavailable
                for name, obj in inspect.getmembers(
                    self.events, lambda obj: isinstance(obj, EventBus)
                ):
                    if name != "status":
                        obj.request_refresh()
            elif event.state == VacuumState.DOCKED:
                self.events.request_refresh(CleanLogEvent)
                self.events.request_refresh(TotalStatsEvent)

        self.events.subscribe(StatusEvent, on_status)

        async def on_stats(_: StatsEvent) -> None:
            self.events.request_refresh(LifeSpanEvent)

        self.events.subscribe(StatsEvent, on_stats)

    async def execute_command(self, command: Union[Command, CustomCommand]) -> None:
        """Execute given command and handle response."""
        if (
            command == Clean(CleanAction.RESUME)
            and self._status.state != VacuumState.PAUSED
        ):
            command = Clean(CleanAction.START)
        elif (
            command == Clean(CleanAction.START)
            and self._status.state == VacuumState.PAUSED
        ):
            command = Clean(CleanAction.RESUME)

        async with self._semaphore:
            response = await self._api_client.send_command(command, self.device_info)

        _LOGGER.debug("Handle command %s: %s", command.name, response)
        if isinstance(command, (CommandWithHandling, CustomCommand)):
            result = command.handle_requested(self.events, response)
            if isinstance(command, CustomCommand):
                # Custom command can be send for implemented commands too.
                # We handle the response explicit to fire event if necessary
                await self.handle_message(command.name, response)

            if result.state == HandlingState.SUCCESS and result.requested_commands:
                # Execute command which are requested by the handler
                tasks = []
                for requested_command in result.requested_commands:
                    tasks.append(
                        asyncio.create_task(self.execute_command(requested_command))
                    )

                await asyncio.gather(*tasks)
        else:
            _LOGGER.warning("Unsupported command! Command %s", command.name)

    def set_available(self, available: bool) -> None:
        """Set available."""
        status = StatusEvent(available, self._status.state)
        self.events.notify(status)

    async def handle_message(
        self, message_name: str, message_data: dict[str, Any]
    ) -> None:
        """Handle the given message.

        :param message_name: message name
        :param message_data: message data
        :return: None
        """
        _LOGGER.debug("Handle message %s: %s", message_name, message_data)
        fw_version = message_data.get("header", {}).get("fwVer", None)
        if fw_version:
            self.fw_version = fw_version

        message_type = MESSAGES.get(message_name, None)
        if message_type:
            message_type.handle(self.events, message_data)
            return

        _LOGGER.debug("Falling back to old handling way for %s", message_name)
        # Handle message starting with "on","off","report" the same as "get" commands
        message_name = re.sub(
            _COMMAND_REPLACE_PATTERN,
            _COMMAND_REPLACE_REPLACEMENT,
            message_name,
        )

        # T8 series and newer
        if message_name.endswith("_V2"):
            message_name = message_name[:-3]

        found_command = MESSAGES.get(
            message_name, COMMANDS_WITH_HANDLING.get(message_name, None)
        )
        if found_command:
            found_command.handle(self.events, message_data)
        else:
            _LOGGER.debug('Unknown message "%s" with %s', message_name, message_data)



from typing import Any, Optional

import pytest

from deebot_client.commands import GetChargeState
from deebot_client.events import StatusEvent
from tests.commands import assert_command_requested
from tests.helpers import get_request_json


@pytest.mark.parametrize(
    "json, expected",
    [
        (get_request_json({"isCharging": 0, "mode": "slot"}), None),
    ],
)
def test_GetChargeState(json: dict[str, Any], expected: Optional[StatusEvent]):
    assert_command_requested(GetChargeState(), json, expected)



"""Life span commands."""
from typing import Any, Union

from ..events import LifeSpan, LifeSpanEvent
from ..message import HandlingResult, HandlingState
from .common import (
    CommandWithHandling,
    CommandWithMqttP2PHandling,
    EventBus,
    _ExecuteCommand,
)


class GetLifeSpan(CommandWithHandling):
    """Get life span command."""

    name = "getLifeSpan"

    def __init__(self) -> None:
        args = [life_span.value for life_span in LifeSpan]
        super().__init__(args)

    @classmethod
    def _handle_body_data_list(cls, event_bus: EventBus, data: list) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        for component in data:
            component_type = LifeSpan(component["type"])
            left = int(component["left"])
            total = int(component["total"])

            if total <= 0:
                raise ValueError("total not positive!")

            percent = round((left / total) * 100, 2)
            event_bus.notify(LifeSpanEvent(component_type, percent, left))

        return HandlingResult.success()


class ResetLifeSpan(_ExecuteCommand, CommandWithMqttP2PHandling):
    """Reset life span command."""

    name = "resetLifeSpan"

    def __init__(
        self, type: Union[str, LifeSpan]  # pylint: disable=redefined-builtin
    ) -> None:
        if isinstance(type, LifeSpan):
            type = type.value

        self._type = type
        super().__init__({"type": type})

    def handle_mqtt_p2p(self, event_bus: EventBus, response: dict[str, Any]) -> None:
        """Handle response received over the mqtt channel "p2p"."""
        result = self.handle(event_bus, response)
        if result.state == HandlingState.SUCCESS:
            event_bus.request_refresh(LifeSpanEvent)



from typing import Any, Optional, Union
from unittest.mock import Mock

from deebot_client.commands import CommandWithHandling, SetCommand
from deebot_client.events import Event
from deebot_client.events.event_bus import EventBus
from deebot_client.message import HandlingState
from tests.helpers import get_message_json


def assert_command_requested(
    command: CommandWithHandling, json: dict[str, Any], expected_event: Optional[Event]
) -> None:
    event_bus = Mock(spec_set=EventBus)

    assert command.name != "invalid"

    result = command.handle_requested(event_bus, json)

    assert result.state == HandlingState.SUCCESS
    if expected_event:
        event_bus.notify.assert_called_once_with(expected_event)
    else:
        event_bus.notify.assert_not_called()


def assert_set_command(
    command: SetCommand,
    args: Union[dict, list, None],
    expected_get_command_event: Event,
) -> None:
    assert command.name != "invalid"
    assert command.args == args

    event_bus = Mock(spec_set=EventBus)

    # Failed to set
    json = {
        "header": {
            "pri": 1,
            "tzm": 480,
            "ts": "1304623069888",
            "ver": "0.0.1",
            "fwVer": "1.8.2",
            "hwVer": "0.1.1",
        },
        "body": {
            "code": 500,
            "msg": "fail",
        },
    }
    command.handle_mqtt_p2p(event_bus, json)
    event_bus.notify.assert_not_called()

    # Success
    command.handle_mqtt_p2p(event_bus, get_message_json(None))
    event_bus.notify.assert_called_once_with(expected_get_command_event)



import pytest

from deebot_client.commands import GetCarpetAutoFanBoost, SetCarpetAutoFanBoost
from deebot_client.events import CarpetAutoFanBoostEvent
from tests.commands import assert_command_requested, assert_set_command
from tests.helpers import get_request_json


@pytest.mark.parametrize("value", [False, True])
def test_get_carpet_auto_fan_boost_requested(value: bool):
    json = get_request_json({"enable": 1 if value else 0})
    assert_command_requested(
        GetCarpetAutoFanBoost(), json, CarpetAutoFanBoostEvent(value)
    )


@pytest.mark.parametrize("value", [False, True])
def test_set_carpet_auto_fan_boost(value: bool):
    args = {"enable": 1 if value else 0}
    assert_set_command(
        SetCarpetAutoFanBoost(value), args, CarpetAutoFanBoostEvent(value)
    )



"""Event constants."""
from typing import Mapping

from ..command import Command
from ..commands import (
    GetAdvancedMode,
    GetBattery,
    GetCachedMapInfo,
    GetCarpetAutoFanBoost,
    GetChargeState,
    GetCleanInfo,
    GetCleanLogs,
    GetContinuousCleaning,
    GetError,
    GetFanSpeed,
    GetLifeSpan,
    GetMajorMap,
    GetMapTrace,
    GetPos,
    GetStats,
    GetTotalStats,
    GetVolume,
    GetWaterInfo,
)
from . import (
    AdvancedModeEvent,
    BatteryEvent,
    CarpetAutoFanBoostEvent,
    CleanLogEvent,
    ContinuousCleaningEvent,
    CustomCommandEvent,
    ErrorEvent,
    Event,
    FanSpeedEvent,
    LifeSpanEvent,
    PositionsEvent,
    ReportStatsEvent,
    RoomsEvent,
    StatsEvent,
    StatusEvent,
    TotalStatsEvent,
    VolumeEvent,
    WaterInfoEvent,
)
from .map import (
    MajorMapEvent,
    MapSetEvent,
    MapSubsetEvent,
    MapTraceEvent,
    MinorMapEvent,
)

EVENT_DTO_REFRESH_COMMANDS: Mapping[type[Event], list[Command]] = {
    AdvancedModeEvent: [GetAdvancedMode()],
    BatteryEvent: [GetBattery()],
    CarpetAutoFanBoostEvent: [GetCarpetAutoFanBoost()],
    CleanLogEvent: [GetCleanLogs()],
    ContinuousCleaningEvent: [GetContinuousCleaning()],
    CustomCommandEvent: [],
    ErrorEvent: [GetError()],
    FanSpeedEvent: [GetFanSpeed()],
    LifeSpanEvent: [GetLifeSpan()],
    MajorMapEvent: [GetMajorMap()],
    MapSetEvent: [],
    MapSubsetEvent: [],
    MapTraceEvent: [GetMapTrace()],
    MinorMapEvent: [],
    PositionsEvent: [GetPos()],
    ReportStatsEvent: [],  # ReportStats cannot be pulled
    RoomsEvent: [GetCachedMapInfo()],
    StatsEvent: [GetStats()],
    StatusEvent: [GetChargeState(), GetCleanInfo()],
    TotalStatsEvent: [GetTotalStats()],
    VolumeEvent: [GetVolume()],
    WaterInfoEvent: [GetWaterInfo()],
}



"""(fan) speed commands."""
from typing import Any, Mapping, Union

from ..events import FanSpeedEvent
from ..message import HandlingResult
from ..util import DisplayNameIntEnum
from .common import EventBus, SetCommand, _NoArgsCommand


class FanSpeedLevel(DisplayNameIntEnum):
    """Enum class for all possible fan speed levels."""

    # Values should be sort from low to high on their meanings
    QUIET = 1000
    NORMAL = 0
    MAX = 1
    MAX_PLUS = 2, "max+"


class GetFanSpeed(_NoArgsCommand):
    """Get fan speed command."""

    name = "getSpeed"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        event_bus.notify(FanSpeedEvent(FanSpeedLevel(int(data["speed"])).display_name))
        return HandlingResult.success()


class SetFanSpeed(SetCommand):
    """Set fan speed command."""

    name = "setSpeed"
    get_command = GetFanSpeed

    def __init__(
        self, speed: Union[str, int, FanSpeedLevel], **kwargs: Mapping[str, Any]
    ) -> None:
        if isinstance(speed, str):
            speed = FanSpeedLevel.get(speed)
        if isinstance(speed, FanSpeedLevel):
            speed = speed.value

        super().__init__({"speed": speed}, **kwargs)



"""Stats messages."""
from typing import Any

from ..events import CleanJobStatus, ReportStatsEvent
from ..events.event_bus import EventBus
from ..message import HandlingResult, Message


class ReportStats(Message):
    """Report stats message."""

    name = "reportStats"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        status = CleanJobStatus.CLEANING

        if data["stop"] != 0:
            status = CleanJobStatus(int(data["stopReason"]))

        stats_event = ReportStatsEvent(
            area=data.get("area"),
            time=data.get("time"),
            type=data.get("type"),
            cleaning_id=data["cid"],
            status=status,
            content=[int(float(x)) for x in data.get("content", "").split(",") if x],
        )
        event_bus.notify(stats_event)
        return HandlingResult.success()



import pytest

from deebot_client.commands import GetBattery
from deebot_client.events import BatteryEvent
from tests.commands import assert_command_requested
from tests.helpers import get_request_json


@pytest.mark.parametrize("percentage", [0, 49, 100])
def test_get_battery_requested(percentage: int):
    json = get_request_json({"value": percentage, "isLow": 1 if percentage < 20 else 0})
    assert_command_requested(GetBattery(), json, BatteryEvent(percentage))



"""Events module."""

from dataclasses import dataclass
from enum import Enum, unique
from typing import Any, Optional

from ..events.base import Event
from ..models import Room, VacuumState
from ..util import DisplayNameIntEnum
from .map import (
    MajorMapEvent,
    MapSetEvent,
    MapSetType,
    MapSubsetEvent,
    MapTraceEvent,
    MinorMapEvent,
    Position,
    PositionsEvent,
    PositionType,
)
from .water_info import WaterAmount, WaterInfoEvent


@dataclass(frozen=True)
class BatteryEvent(Event):
    """Battery event representation."""

    value: int


class CleanJobStatus(DisplayNameIntEnum):
    """Enum of the different clean job status."""

    CLEANING = -1
    # below the identified stop_reason values
    FINISHED = 1
    MANUAL_STOPPED = 2, "manual stopped"
    FINISHED_WITH_WARNINGS = 3, "finished with warnings"


@dataclass(frozen=True)
class CleanLogEntry:
    """Clean log entry representation."""

    timestamp: int
    image_url: str
    type: str
    area: int
    stop_reason: CleanJobStatus
    duration: int  # in seconds


@dataclass(frozen=True)
class CleanLogEvent(Event):
    """Clean log event representation."""

    logs: list[CleanLogEntry]


@dataclass(frozen=True)
class CustomCommandEvent(Event):
    """Custom command event representation."""

    name: str
    response: dict[str, Any]


@dataclass(frozen=True)
class ErrorEvent(Event):
    """Error event representation."""

    code: int
    description: Optional[str]


@dataclass(frozen=True)
class FanSpeedEvent(Event):
    """Fan speed event representation."""

    speed: str


@unique
class LifeSpan(str, Enum):
    """Enum class for all possible life span components."""

    SIDE_BRUSH = "sideBrush"
    BRUSH = "brush"
    FILTER = "heap"


@dataclass(frozen=True)
class LifeSpanEvent(Event):
    """Life span event representation."""

    type: LifeSpan
    percent: float
    remaining: int  # in minutes


@dataclass(frozen=True)
class RoomsEvent(Event):
    """Room event representation."""

    rooms: list[Room]


@dataclass(frozen=True)
class StatsEvent(Event):
    """Stats event representation."""

    area: Optional[int]
    time: Optional[int]
    type: Optional[str]


@dataclass(frozen=True)
class ReportStatsEvent(StatsEvent):
    """Report stats event representation."""

    cleaning_id: str
    status: CleanJobStatus
    content: Optional[list[int]]


@dataclass(frozen=True)
class TotalStatsEvent(Event):
    """Total stats event representation."""

    area: int
    time: int
    cleanings: int


@dataclass(frozen=True)
class StatusEvent(Event):
    """Status event representation."""

    available: bool
    state: Optional[VacuumState]


@dataclass(frozen=True)
class VolumeEvent(Event):
    """Volume event."""

    volume: int
    maximum: Optional[int]


@dataclass(frozen=True)
class EnableEvent(Event):
    """Enabled event."""

    enable: bool


@dataclass(frozen=True)
class AdvancedModeEvent(EnableEvent):
    """Advanced mode event."""


@dataclass(frozen=True)
class ContinuousCleaningEvent(EnableEvent):
    """Continuous cleaning event."""


@dataclass(frozen=True)
class CarpetAutoFanBoostEvent(EnableEvent):
    """Carpet pressure event."""



"""Clean commands."""
from enum import Enum, unique
from typing import Any, Optional

from ..events import StatusEvent
from ..logging_filter import get_logger
from ..message import HandlingResult
from ..models import VacuumState
from .common import EventBus, _ExecuteCommand, _NoArgsCommand

_LOGGER = get_logger(__name__)


@unique
class CleanAction(str, Enum):
    """Enum class for all possible clean actions."""

    START = "start"
    PAUSE = "pause"
    RESUME = "resume"
    STOP = "stop"


@unique
class CleanMode(str, Enum):
    """Enum class for all possible clean modes."""

    AUTO = "auto"
    SPOT_AREA = "spotArea"
    CUSTOM_AREA = "customArea"


class Clean(_ExecuteCommand):
    """Clean command."""

    name = "clean"

    def __init__(self, action: CleanAction) -> None:
        args = {"act": action.value}
        if action == CleanAction.START:
            args["type"] = CleanMode.AUTO.value
        super().__init__(args)


class CleanArea(Clean):
    """Clean area command."""

    def __init__(self, mode: CleanMode, area: str, cleanings: int = 1) -> None:
        super().__init__(CleanAction.START)
        if not isinstance(self.args, dict):
            raise ValueError("args must be a dict!")

        self.args["type"] = mode.value
        self.args["content"] = str(area)
        self.args["count"] = cleanings


class GetCleanInfo(_NoArgsCommand):
    """Get clean info command."""

    name = "getCleanInfo"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """

        status: Optional[VacuumState] = None
        state = data.get("state")
        if data.get("trigger") == "alert":
            status = VacuumState.ERROR
        elif state == "clean":
            clean_state = data.get("cleanState", {})
            motion_state = clean_state.get("motionState")
            if motion_state == "working":
                status = VacuumState.CLEANING
            elif motion_state == "pause":
                status = VacuumState.PAUSED
            elif motion_state == "goCharging":
                status = VacuumState.RETURNING

            clean_type = clean_state.get("type")
            content = clean_state.get("content", {})
            if "type" in content:
                clean_type = content.get("type")

            if clean_type == "customArea":
                area_values = content
                if "value" in content:
                    area_values = content.get("value")

                _LOGGER.debug("Last custom area values (x1,y1,x2,y2): %s", area_values)

        elif state == "goCharging":
            status = VacuumState.RETURNING
        elif state == "idle":
            status = VacuumState.IDLE

        if status:
            event_bus.notify(StatusEvent(True, status))
            return HandlingResult.success()

        return HandlingResult.analyse()



"""Water info commands."""
from typing import Any, Mapping, Union

from ..events import WaterAmount, WaterInfoEvent
from ..message import HandlingResult
from .common import EventBus, SetCommand, _NoArgsCommand


class GetWaterInfo(_NoArgsCommand):
    """Get water info command."""

    name = "getWaterInfo"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        # todo enable can be missing pylint: disable=fixme
        mop_attached = bool(data.get("enable"))

        event_bus.notify(WaterInfoEvent(mop_attached, WaterAmount(int(data["amount"]))))
        return HandlingResult.success()


class SetWaterInfo(SetCommand):
    """Set water info command."""

    name = "setWaterInfo"
    get_command = GetWaterInfo

    def __init__(
        self, amount: Union[str, int, WaterAmount], **kwargs: Mapping[str, Any]
    ) -> None:
        # removing "enable" as we don't can set it
        kwargs.pop("enable", None)

        if isinstance(amount, str):
            amount = WaterAmount.get(amount)
        if isinstance(amount, WaterAmount):
            amount = amount.value

        super().__init__({"amount": amount}, **kwargs)



"""Map event module."""
from dataclasses import dataclass
from enum import Enum, unique
from typing import Any, Optional

from ..events import Event


@unique
class PositionType(str, Enum):
    """Position type enum."""

    DEEBOT = "deebotPos"
    CHARGER = "chargePos"


@dataclass(frozen=True)
class Position:
    """Position representation."""

    type: PositionType
    x: int
    y: int


@dataclass(frozen=True)
class PositionsEvent(Event):
    """Position event representation."""

    positions: list[Position]


@dataclass(frozen=True)
class MapTraceEvent(Event):
    """Map trace event representation."""

    start: int
    total: int
    data: str


@dataclass(frozen=True)
class MajorMapEvent(Event):
    """Major map event."""

    requested: bool
    map_id: str
    values: list[str]


@dataclass(frozen=True)
class MinorMapEvent(Event):
    """Minor map event."""

    index: int
    value: str


@unique
class MapSetType(str, Enum):
    """Map set type enum."""

    ROOMS = "ar"
    VIRTUAL_WALLS = "vw"
    NO_MOP_ZONES = "mw"

    @classmethod
    def has_value(cls, value: Any) -> bool:
        """Check if value exists."""

        return value in cls._value2member_map_  # pylint: disable=no-member


@dataclass(frozen=True)
class MapSetEvent(Event):
    """Map set event."""

    type: MapSetType
    subsets: list[int]


@dataclass(frozen=True)
class MapSubsetEvent(Event):
    """Map subset event."""

    id: int
    type: MapSetType
    coordinates: str
    subtype: Optional[str] = None



"""Base messages."""
import functools
from abc import ABC
from dataclasses import dataclass
from enum import IntEnum, auto
from typing import Any, Callable, Optional, Union, final

from .events.event_bus import EventBus
from .logging_filter import get_logger

_LOGGER = get_logger(__name__)


class HandlingState(IntEnum):
    """Handling state enum."""

    SUCCESS = auto()
    FAILED = auto()
    ERROR = auto()
    ANALYSE = auto()
    ANALYSE_LOGGED = auto()


@dataclass(frozen=True)
class HandlingResult:
    """Handling result object."""

    state: HandlingState
    args: Optional[dict[str, Any]] = None

    @classmethod
    def success(cls) -> "HandlingResult":
        """Create result with handling success."""
        return HandlingResult(HandlingState.SUCCESS)

    @classmethod
    def analyse(cls) -> "HandlingResult":
        """Create result with handling analyse."""
        return HandlingResult(HandlingState.ANALYSE)


def _handle_error_or_analyse(
    func: Callable[[type["Message"], EventBus, dict[str, Any]], HandlingResult]
) -> Callable[[type["Message"], EventBus, dict[str, Any]], HandlingResult]:
    """Handle error or None response."""

    @functools.wraps(func)
    def wrapper(
        cls: type["Message"], event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        try:
            response = func(cls, event_bus, data)
            if response.state == HandlingState.ANALYSE:
                _LOGGER.debug("Could not handle %s message: %s", cls.name, data)
                return HandlingResult(HandlingState.ANALYSE_LOGGED, response.args)
            return response
        except Exception:  # pylint: disable=broad-except
            _LOGGER.warning("Could not parse %s: %s", cls.name, data, exc_info=True)
            return HandlingResult(HandlingState.ERROR)

    return wrapper


class Message(ABC):
    """Message with handling code."""

    # will be overwritten in subclasses
    name = "__invalid__"

    @classmethod
    def _handle_body_data_list(cls, event_bus: EventBus, data: list) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        raise NotImplementedError

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        raise NotImplementedError

    @classmethod
    def _handle_body_data(
        cls, event_bus: EventBus, data: Union[dict[str, Any], list]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        if isinstance(data, dict):
            return cls._handle_body_data_dict(event_bus, data)

        if isinstance(data, list):
            return cls._handle_body_data_list(event_bus, data)

    @classmethod
    @_handle_error_or_analyse
    @final
    def __handle_body_data(
        cls, event_bus: EventBus, data: Union[dict[str, Any], list]
    ) -> HandlingResult:
        return cls._handle_body_data(event_bus, data)

    @classmethod
    def _handle_body(cls, event_bus: EventBus, body: dict[str, Any]) -> HandlingResult:
        """Handle message->body and notify the correct event subscribers.

        :return: A message response
        """
        data = body.get("data", body)
        return cls.__handle_body_data(event_bus, data)

    @classmethod
    @_handle_error_or_analyse
    @final
    def __handle_body(cls, event_bus: EventBus, body: dict[str, Any]) -> HandlingResult:
        return cls._handle_body(event_bus, body)

    @classmethod
    @_handle_error_or_analyse
    @final
    def handle(cls, event_bus: EventBus, message: dict[str, Any]) -> HandlingResult:
        """Handle message and notify the correct event subscribers.

        :return: A message response
        """
        data_body = message.get("body", message)
        return cls.__handle_body(event_bus, data_body)



"""Constatns module."""

REALM = "ecouser.net"
PATH_API_APPSVR_APP = "appsvr/app.do"
PATH_API_PIM_PRODUCT_IOT_MAP = "pim/product/getProductIotMap"
PATH_API_IOT_DEVMANAGER = "iot/devmanager.do"
PATH_API_LG_LOG = "lg/log.do"



"""Internal api client module."""
import asyncio
from typing import Any, Optional
from urllib.parse import urljoin

from aiohttp import ClientResponseError

from .const import REALM
from .logging_filter import get_logger
from .models import Configuration, Credentials

_LOGGER = get_logger(__name__)


def _get_portal_url(config: Configuration, path: str) -> str:
    subdomain = f"portal-{config.continent}" if config.country != "cn" else "portal"
    return urljoin(f"https://{subdomain}.ecouser.net/api/", path)


class _InternalApiClient:
    """Internal api client.

    Only required for AuthClient and ApiClient. For all other usecases use APIClient instead.
    """

    def __init__(self, config: Configuration):
        self._config = config

    async def post(
        self,
        path: str,
        json: dict[str, Any],
        *,
        query_params: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, Any]] = None,
        credentials: Optional[Credentials] = None,
    ) -> dict[str, Any]:
        """Perform a post request."""
        url = _get_portal_url(self._config, path)

        _LOGGER.debug("calling api %s with %s", path, json)

        if credentials is not None:
            json.update(
                {
                    "auth": {
                        "with": "users",
                        "userid": credentials.user_id,
                        "realm": REALM,
                        "token": credentials.token,
                        "resource": self._config.device_id,
                    }
                }
            )

        try:
            async with self._config.session.post(
                url,
                json=json,
                params=query_params,
                headers=headers,
                timeout=60,
                ssl=self._config.verify_ssl,
            ) as res:
                res.raise_for_status()
                if res.status != 200:
                    _LOGGER.warning("Error calling API (%d): %s", res.status, path)
                    return {}

                response_data: dict[str, Any] = await res.json()
                _LOGGER.debug("got %s", response_data)
                return response_data
        except asyncio.TimeoutError:
            command = ""
            if "cmdName" in json:
                command = f"({json['cmdName']})"
            _LOGGER.warning("Timeout reached on api path: %s%s", path, command)
        except ClientResponseError as err:
            if err.status == 502:
                _LOGGER.info(
                    "Error calling API (502): Unfortunately the ecovacs api is unreliable. URL was: %s",
                    url,
                )
            else:
                _LOGGER.warning("Error calling API (%sd): %s", err.status, url)

        return {}



"""Charge commands."""
from typing import Any

from ..events import StatusEvent
from ..logging_filter import get_logger
from ..message import HandlingResult
from ..models import VacuumState
from .common import EventBus, _ExecuteCommand
from .const import CODE

_LOGGER = get_logger(__name__)


class Charge(_ExecuteCommand):
    """Charge command."""

    name = "charge"

    def __init__(self) -> None:
        super().__init__({"act": "go"})

    @classmethod
    def _handle_body(cls, event_bus: EventBus, body: dict[str, Any]) -> HandlingResult:
        """Handle message->body and notify the correct event subscribers.

        :return: A message response
        """
        code = int(body.get(CODE, -1))
        if code == 0:
            event_bus.notify(StatusEvent(True, VacuumState.RETURNING))
            return HandlingResult.success()

        if code == 30007:
            # bot is already charging
            event_bus.notify(StatusEvent(True, VacuumState.DOCKED))
            return HandlingResult.success()

        return super()._handle_body(event_bus, body)



from typing import Any

import pytest

from deebot_client.commands import Charge
from deebot_client.events import StatusEvent
from deebot_client.models import VacuumState
from tests.commands import assert_command_requested
from tests.helpers import get_request_json


def prepare_json_docked_test():
    json = get_request_json(None)
    json["resp"]["body"]["code"] = 30007
    return json


@pytest.mark.parametrize(
    "json, expected",
    [
        (get_request_json(None), StatusEvent(True, VacuumState.RETURNING)),
        (prepare_json_docked_test(), StatusEvent(True, VacuumState.DOCKED)),
    ],
)
def test_charge(json: dict[str, Any], expected: StatusEvent):
    assert_command_requested(Charge(), json, expected)



"""Base commands."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Mapping, Optional, Union

from ..command import Command
from ..events import EnableEvent
from ..events.event_bus import EventBus
from ..logging_filter import get_logger
from ..message import HandlingResult, HandlingState, Message
from .const import CODE

_LOGGER = get_logger(__name__)


@dataclass(frozen=True)
class CommandResult(HandlingResult):
    """Command result object."""

    requested_commands: Optional[list[Command]] = None

    @classmethod
    def success(cls) -> "CommandResult":
        """Create result with handling success."""
        return CommandResult(HandlingState.SUCCESS)

    @classmethod
    def analyse(cls) -> "CommandResult":
        """Create result with handling analyse."""
        return CommandResult(HandlingState.ANALYSE)


class CommandWithHandling(Command, Message, ABC):
    """Command, which handle response by itself."""

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    def handle_requested(
        self, event_bus: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        if response.get("ret") == "ok":
            data = response.get("resp", response)
            result = self.handle(event_bus, data)
            return CommandResult(result.state, result.args)

        _LOGGER.warning('Command "%s" was not successfully: %s', self.name, response)
        return CommandResult(HandlingState.FAILED)


class CommandWithMqttP2PHandling(CommandWithHandling, ABC):
    """Command which handles also mqtt p2p messages."""

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    def handle_mqtt_p2p(self, event_bus: EventBus, response: dict[str, Any]) -> None:
        """Handle response received over the mqtt channel "p2p"."""
        raise NotImplementedError


class _NoArgsCommand(CommandWithHandling, ABC):
    """Command without args."""

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    def __init__(self) -> None:
        super().__init__()


class _ExecuteCommand(CommandWithHandling, ABC):
    """Command, which is executing something (ex. Charge)."""

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    @classmethod
    def _handle_body(cls, event_bus: EventBus, body: dict[str, Any]) -> HandlingResult:
        """Handle message->body and notify the correct event subscribers.

        :return: A message response
        """
        # Success event looks like { "code": 0, "msg": "ok" }
        if body.get(CODE, -1) == 0:
            return HandlingResult.success()

        _LOGGER.warning('Command "%s" was not successfully. body=%s', cls.name, body)
        return HandlingResult(HandlingState.FAILED)


class SetCommand(_ExecuteCommand, CommandWithMqttP2PHandling, ABC):
    """Base set command.

    Command needs to be linked to the "get" command, for handling (updating) the sensors.
    """

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    def __init__(
        self,
        args: Union[dict, list, None],
        **kwargs: Mapping[str, Any],
    ) -> None:
        if kwargs:
            _LOGGER.debug("Following passed parameters will be ignored: %s", kwargs)

        super().__init__(args)

    @property
    @abstractmethod
    def get_command(self) -> type[CommandWithHandling]:
        """Return the corresponding "get" command."""
        raise NotImplementedError

    def handle_mqtt_p2p(self, event_bus: EventBus, response: dict[str, Any]) -> None:
        """Handle response received over the mqtt channel "p2p"."""
        result = self.handle(event_bus, response)
        if result.state == HandlingState.SUCCESS and isinstance(self.args, dict):
            self.get_command.handle(event_bus, self.args)


class _GetEnableCommand(_NoArgsCommand):
    """Abstract get enable command."""

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    @classmethod
    @property
    @abstractmethod
    def event_type(cls) -> type[EnableEvent]:
        """Event type."""
        raise NotImplementedError

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        event: EnableEvent = cls.event_type(bool(data["enable"]))  # type: ignore
        event_bus.notify(event)
        return HandlingResult.success()


class SetEnableCommand(SetCommand):
    """Abstract set enable command."""

    # required as name is class variable, will be overwritten in subclasses
    name = "__invalid__"

    def __init__(self, enable: Union[int, bool], **kwargs: Mapping[str, Any]) -> None:
        if isinstance(enable, bool):
            enable = 1 if enable else 0
        super().__init__({"enable": enable}, **kwargs)



from typing import Union

import pytest

from deebot_client.commands import FanSpeedLevel, GetFanSpeed, SetFanSpeed
from deebot_client.events import FanSpeedEvent
from tests.commands import assert_command_requested
from tests.helpers import get_request_json, verify_DisplayNameEnum_unique


def test_FanSpeedLevel_unique():
    verify_DisplayNameEnum_unique(FanSpeedLevel)


def test_GetFanSpeed_requested():
    json = get_request_json({"speed": 2})
    assert_command_requested(GetFanSpeed(), json, FanSpeedEvent("max+"))


@pytest.mark.parametrize(
    "value, expected", [("quiet", 1000), ("max+", 2), (0, 0), (FanSpeedLevel.MAX, 1)]
)
def test_SetFanSpeed(value: Union[str, int, FanSpeedLevel], expected: int):

    command = SetFanSpeed(value)
    assert command.name == "setSpeed"
    assert command.args == {"speed": expected}



"""Commands module."""
from typing import Dict, Type

from .advanced_mode import GetAdvancedMode, SetAdvancedMode
from .battery import GetBattery
from .carpet import GetCarpetAutoFanBoost, SetCarpetAutoFanBoost
from .charge import Charge
from .charge_state import GetChargeState
from .clean import Clean, CleanArea, GetCleanInfo
from .clean_logs import GetCleanLogs
from .common import CommandWithHandling, CommandWithMqttP2PHandling, SetCommand
from .continuous_cleaning import GetContinuousCleaning, SetContinuousCleaning
from .error import GetError
from .fan_speed import FanSpeedLevel, GetFanSpeed, SetFanSpeed
from .life_span import GetLifeSpan, ResetLifeSpan
from .map import (
    GetCachedMapInfo,
    GetMajorMap,
    GetMapSet,
    GetMapSubSet,
    GetMapTrace,
    GetMinorMap,
)
from .play_sound import PlaySound
from .pos import GetPos
from .relocation import SetRelocationState
from .stats import GetStats, GetTotalStats
from .volume import GetVolume, SetVolume
from .water_info import GetWaterInfo, SetWaterInfo

# fmt: off
# ordered by file asc
_COMMANDS: list[type[CommandWithHandling]] = [
    GetAdvancedMode,
    SetAdvancedMode,

    GetBattery,

    GetCarpetAutoFanBoost,
    SetCarpetAutoFanBoost,

    Charge,

    GetChargeState,

    Clean,
    CleanArea,
    GetCleanInfo,

    GetCleanLogs,

    GetContinuousCleaning,
    SetContinuousCleaning,

    GetError,

    GetFanSpeed,
    SetFanSpeed,

    GetLifeSpan,
    ResetLifeSpan,

    GetCachedMapInfo,
    GetMajorMap,
    GetMapSet,
    GetMapSubSet,
    GetMapTrace,
    GetMinorMap,

    PlaySound,

    GetPos,

    SetRelocationState,

    GetStats,
    GetTotalStats,

    GetVolume,
    SetVolume,

    GetWaterInfo,
    SetWaterInfo,
]
# fmt: on

COMMANDS_WITH_HANDLING: dict[str, type[CommandWithHandling]] = {
    cmd.name: cmd for cmd in _COMMANDS
}

COMMANDS_WITH_MQTT_P2P_HANDLING: dict[str, type[CommandWithMqttP2PHandling]] = {
    cmd_name: cmd
    for (cmd_name, cmd) in COMMANDS_WITH_HANDLING.items()
    if issubclass(cmd, CommandWithMqttP2PHandling)
}



"""Deebot client module."""
from typing import Tuple

from ._api_client import _InternalApiClient
from .api_client import ApiClient
from .authentication import Authenticator
from .models import Configuration


def create_instances(
    config: Configuration, account_id: str, password_hash: str
) -> tuple[Authenticator, ApiClient]:
    """Create a authenticator and api client instance."""
    internal_api_client = _InternalApiClient(config)
    authenticator = Authenticator(
        config, internal_api_client, account_id, password_hash
    )
    api_client = ApiClient(internal_api_client, authenticator)

    return authenticator, api_client



from typing import Any

import pytest

from deebot_client.commands import GetCleanInfo
from deebot_client.events import StatusEvent
from deebot_client.models import VacuumState
from tests.commands import assert_command_requested
from tests.helpers import get_request_json


@pytest.mark.parametrize(
    "json, expected",
    [
        (
            get_request_json({"trigger": "none", "state": "idle"}),
            StatusEvent(True, VacuumState.IDLE),
        ),
    ],
)
def test_GetCleanInfo(json: dict[str, Any], expected: StatusEvent):
    assert_command_requested(GetCleanInfo(), json, expected)



"""Api client module."""
from datetime import datetime
from typing import Any, Union
from urllib.parse import urljoin

from ._api_client import _InternalApiClient
from .authentication import Authenticator
from .command import Command
from .commands import GetCleanLogs
from .commands.custom import CustomCommand
from .const import (
    PATH_API_APPSVR_APP,
    PATH_API_IOT_DEVMANAGER,
    PATH_API_LG_LOG,
    PATH_API_PIM_PRODUCT_IOT_MAP,
)
from .exceptions import ApiError
from .logging_filter import get_logger
from .models import Configuration, DeviceInfo

_LOGGER = get_logger(__name__)
_REQUEST_HEADERS = {
    "User-Agent": "Dalvik/2.1.0 (Linux; U; Android 5.1.1; A5010 Build/LMY48Z)",
}


def _get_portal_url(config: Configuration, path: str) -> str:
    subdomain = f"portal-{config.continent}" if config.country != "cn" else "portal"
    return urljoin(f"https://{subdomain}.ecouser.net/api/", path)


class ApiClient:
    """Api client."""

    def __init__(
        self, internal_api_client: _InternalApiClient, authenticator: Authenticator
    ):
        self._api_client = internal_api_client
        self._authenticator = authenticator

    async def get_devices(self) -> list[DeviceInfo]:
        """Get compatible devices."""
        credentials = await self._authenticator.authenticate()
        json = {
            "userid": credentials.user_id,
            "todo": "GetGlobalDeviceList",
        }
        resp = await self._api_client.post(
            PATH_API_APPSVR_APP, json, credentials=credentials
        )

        if resp.get("code", None) == 0:
            devices: list[DeviceInfo] = []
            for device in resp["devices"]:
                if device.get("company") == "eco-ng":
                    devices.append(DeviceInfo(device))
                else:
                    _LOGGER.debug("Skipping device as it is not supported: %s", device)
            return devices
        _LOGGER.error("Failed to get devices")
        raise ApiError(f"failure {resp['error']} ({resp['errno']}) on getting devices")

    async def get_product_iot_map(self) -> dict[str, Any]:
        """Get product iot map."""
        resp = await self._api_client.post(
            PATH_API_PIM_PRODUCT_IOT_MAP,
            {},
            credentials=await self._authenticator.authenticate(),
        )

        if resp.get("code", None) in [0, "0000"]:
            result: dict[str, Any] = {}
            for entry in resp["data"]:
                result[entry["classid"]] = entry["product"]
            return result
        _LOGGER.error("Failed to get product iot map")
        raise ApiError(
            f"failure {resp['error']} ({resp['errno']}) on getting product iot map"
        )

    async def send_command(
        self,
        command: Union[Command, CustomCommand],
        device_info: DeviceInfo,
    ) -> dict[str, Any]:
        """Send json command for given vacuum to the api."""
        query_params = {}
        json: dict[str, Any]

        if command.name == GetCleanLogs.name:
            json = {
                "td": command.name,
                "did": device_info.did,
                "resource": device_info.resource,
            }

            path = PATH_API_LG_LOG
        else:
            payload = {
                "header": {
                    "pri": "1",
                    "ts": datetime.now().timestamp(),
                    "tzm": 480,
                    "ver": "0.0.50",
                }
            }

            if len(command.args) > 0:
                payload["body"] = {"data": command.args}

            json = {
                "cmdName": command.name,
                "payload": payload,
                "payloadType": "j",
                "td": "q",
                "toId": device_info.did,
                "toRes": device_info.resource,
                "toType": device_info.get_class,
            }

            path = PATH_API_IOT_DEVMANAGER
            query_params.update({"mid": json["toType"], "did": json["toId"]})

        credentials = await self._authenticator.authenticate()
        query_params.update(
            {
                "td": json["td"],
                "u": credentials.user_id,
                "cv": "1.67.3",
                "t": "a",
                "av": "1.3.1",
            }
        )

        return await self._api_client.post(
            path,
            json,
            query_params=query_params,
            headers=_REQUEST_HEADERS,
            credentials=credentials,
        )



"""Carpet pressure command module."""

from ..events import CarpetAutoFanBoostEvent
from .common import SetEnableCommand, _GetEnableCommand


class GetCarpetAutoFanBoost(_GetEnableCommand):
    """Get carpet auto fan boost command."""

    name = "getCarpertPressure"
    event_type = CarpetAutoFanBoostEvent


class SetCarpetAutoFanBoost(SetEnableCommand):
    """Set carpet auto fan boost command."""

    name = "setCarpertPressure"
    get_command = GetCarpetAutoFanBoost



"""Error commands."""
from typing import Any

from ..events import ErrorEvent, StatusEvent
from ..message import HandlingResult
from ..models import VacuumState
from .common import EventBus, _NoArgsCommand


class GetError(_NoArgsCommand):
    """Get error command."""

    name = "getError"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        codes = data.get("code", [])
        if codes:
            # the last error code
            error = codes[-1]

            if error is not None:
                description = _ERROR_CODES.get(error)
                if error != 0:
                    event_bus.notify(StatusEvent(True, VacuumState.ERROR))
                event_bus.notify(ErrorEvent(error, description))
                return HandlingResult.success()

        return HandlingResult.analyse()


# from https://github.com/mrbungle64/ecovacs-deebot.js/blob/master/library/errorCodes.js
_ERROR_CODES = {
    -3: "Error parsing response data",
    -2: "Internal error",
    -1: "Host not reachable or communication malfunction",
    0: "NoError: Robot is operational",
    3: "RequestOAuthError: Authentication error",
    7: "log data is not found",
    100: "NoError: Robot is operational",
    101: "BatteryLow: Low battery",
    102: "HostHang: Robot is off the floor",
    103: "WheelAbnormal: Driving Wheel malfunction",
    104: "DownSensorAbnormal: Excess dust on the Anti-Drop Sensors",
    105: "Stuck: Robot is stuck",
    106: "SideBrushExhausted: Side Brushes have expired",
    107: "DustCaseHeapExhausted: Dust case filter expired",
    108: "SideAbnormal: Side Brushes are tangled",
    109: "RollAbnormal: Main Brush is tangled",
    110: "NoDustBox: Dust Bin Not installed",
    111: "BumpAbnormal: Bump sensor stuck",
    112: 'LDS: LDS "Laser Distance Sensor" malfunction',
    113: "MainBrushExhausted: Main brush has expired",
    114: "DustCaseFilled: Dust bin full",
    115: "BatteryError:",
    116: "ForwardLookingError:",
    117: "GyroscopeError:",
    118: "StrainerBlock:",
    119: "FanError:",
    120: "WaterBoxError:",
    201: "AirFilterUninstall:",
    202: "UltrasonicComponentAbnormal",
    203: "SmallWheelError",
    204: "WheelHang",
    205: "IonSterilizeExhausted",
    206: "IonSterilizeAbnormal",
    207: "IonSterilizeFault",
    404: "Recipient unavailable",
    500: "Request Timeout",
    601: "ERROR_ClosedAIVISideAbnormal",
    602: "ClosedAIVIRollAbnormal",
}



"""Authentication module."""
import asyncio
import time
from typing import Any, Callable, Mapping, Optional, Union

from aiohttp import hdrs

from ._api_client import _InternalApiClient
from .const import REALM
from .exceptions import AuthenticationError, InvalidAuthenticationError
from .logging_filter import get_logger
from .models import Configuration, Credentials
from .util import md5

_LOGGER = get_logger(__name__)

_CLIENT_KEY = "1520391301804"
_CLIENT_SECRET = "6c319b2a5cd3e66e39159c2e28f2fce9"
_AUTH_CLIENT_KEY = "1520391491841"
_AUTH_CLIENT_SECRET = "77ef58ce3afbe337da74aa8c5ab963a9"
_USER_LOGIN_URL_FORMAT = (
    "https://gl-{country}-api.ecovacs.{tld}/v1/private/{country}/{lang}/{deviceId}/{appCode}/"
    "{appVersion}/{channel}/{deviceType}/user/login"
)
_GLOBAL_AUTHCODE_URL_FORMAT = (
    "https://gl-{country}-openapi.ecovacs.{tld}/v1/global/auth/getAuthCode"
)
_PATH_USERS_USER = "users/user.do"
_META = {
    "lang": "EN",
    "appCode": "global_e",
    "appVersion": "1.6.3",
    "channel": "google_play",
    "deviceType": "1",
}


class _AuthClient:
    """Ecovacs auth client."""

    def __init__(
        self,
        config: Configuration,
        internal_api_client: _InternalApiClient,
        account_id: str,
        password_hash: str,
    ):
        self._config = config
        self._api_client = internal_api_client
        self._account_id = account_id
        self._password_hash = password_hash
        self._tld = "com" if self._config.country != "cn" else "cn"

        self._meta: dict[str, str] = {
            **_META,
            "country": self._config.country,
            "deviceId": self._config.device_id,
        }

    async def login(self) -> Credentials:
        """Login using username and password."""
        _LOGGER.debug("Start login to EcovacsAPI")
        login_password_resp = await self.__call_login_api(
            self._account_id, self._password_hash
        )
        user_id = login_password_resp["uid"]

        auth_code = await self.__call_auth_api(
            login_password_resp["accessToken"], user_id
        )

        login_token_resp = await self.__call_login_by_it_token(user_id, auth_code)
        if login_token_resp["userId"] != user_id:
            _LOGGER.debug("Switching to shorter UID")
            user_id = login_token_resp["userId"]

        user_access_token = login_token_resp["token"]
        # last is validity in milliseconds. Usually 7 days
        # we set the expiry at 99% of the validity
        # 604800 = 7 days
        expires_at = int(
            time.time() + int(login_token_resp.get("last", 604800)) / 1000 * 0.99
        )

        _LOGGER.debug("Login to EcovacsAPI successfully")
        return Credentials(
            token=user_access_token,
            user_id=user_id,
            expires_at=expires_at,
        )

    async def __do_auth_response(
        self, url: str, params: dict[str, Any]
    ) -> dict[str, Any]:
        async with self._config.session.get(
            url, params=params, timeout=60, ssl=self._config.verify_ssl
        ) as res:
            res.raise_for_status()

            # ecovacs returns a json but content_type header is set to text
            content_type = res.headers.get(hdrs.CONTENT_TYPE, "").lower()
            json = await res.json(content_type=content_type)
            _LOGGER.debug("got %s", json)
            # todo better error handling # pylint: disable=fixme
            if json["code"] == "0000":
                data: dict[str, Any] = json["data"]
                return data
            if json["code"] in ["1005", "1010"]:
                raise InvalidAuthenticationError

            _LOGGER.error("call to %s failed with %s", url, json)
            raise AuthenticationError(
                f"failure code {json['code']} ({json['msg']}) for call {url}"
            )

    async def __call_login_api(
        self, account_id: str, password_hash: str
    ) -> dict[str, Any]:
        _LOGGER.debug("calling login api")
        params: dict[str, Union[str, int]] = {
            "account": account_id,
            "password": password_hash,
            "requestId": md5(str(time.time())),
            "authTimespan": int(time.time() * 1000),
            "authTimeZone": "GMT-8",
        }

        url = _USER_LOGIN_URL_FORMAT.format(**self._meta, tld=self._tld)

        if self._config.country.lower() == "cn":
            url += "CheckMobile"

        return await self.__do_auth_response(
            url, self.__sign(params, self._meta, _CLIENT_KEY, _CLIENT_SECRET)
        )

    @staticmethod
    def __sign(
        params: dict[str, Union[str, int]],
        additional_sign_params: Mapping[str, Union[str, int]],
        key: str,
        secret: str,
    ) -> dict[str, Union[str, int]]:
        sign_data: dict[str, Union[str, int]] = {**additional_sign_params, **params}
        sign_on_text = (
            key
            + "".join([k + "=" + str(sign_data[k]) for k in sorted(sign_data.keys())])
            + secret
        )
        params["authSign"] = md5(sign_on_text)
        params["authAppkey"] = key
        return params

    async def __call_auth_api(self, access_token: str, user_id: str) -> str:
        _LOGGER.debug("calling auth api")
        params: dict[str, Union[str, int]] = {
            "uid": user_id,
            "accessToken": access_token,
            "bizType": "ECOVACS_IOT",
            "deviceId": self._meta["deviceId"],
            "authTimespan": int(time.time() * 1000),
        }

        url = _GLOBAL_AUTHCODE_URL_FORMAT.format(**self._meta, tld=self._tld)

        res = await self.__do_auth_response(
            url,
            self.__sign(
                params, {"openId": "global"}, _AUTH_CLIENT_KEY, _AUTH_CLIENT_SECRET
            ),
        )
        return str(res["authCode"])

    async def __call_login_by_it_token(
        self, user_id: str, auth_code: str
    ) -> dict[str, str]:
        data = {
            "edition": "ECOGLOBLE",
            "userId": user_id,
            "token": auth_code,
            "realm": REALM,
            "resource": self._config.device_id,
            "org": "ECOWW" if self._config.country != "cn" else "ECOCN",
            "last": "",
            "country": self._config.country.upper()
            if self._config.country != "cn"
            else "Chinese",
            "todo": "loginByItToken",
        }

        for i in range(3):
            resp = await self._api_client.post(_PATH_USERS_USER, data)
            if resp["result"] == "ok":
                return resp
            if resp["result"] == "fail" and resp["error"] == "set token error.":
                # If it is a set token error try again
                _LOGGER.warning("loginByItToken set token error, attempt %d/3", i + 2)
                continue

            _LOGGER.error("call to %s failed with %s", _PATH_USERS_USER, resp)
            raise AuthenticationError(
                f"failure {resp['error']} ({resp['errno']}) for call {_PATH_USERS_USER}"
            )

        raise AuthenticationError("failed to login with token")


class Authenticator:
    """Authenticator."""

    def __init__(
        self,
        config: Configuration,
        ecovacs_api_client: _InternalApiClient,
        account_id: str,
        password_hash: str,
    ):
        self._auth_client = _AuthClient(
            config,
            ecovacs_api_client,
            account_id,
            password_hash,
        )

        self._lock = asyncio.Lock()
        self._on_credentials_changed: set[Callable[[Credentials], None]] = set()
        self._credentials: Optional[Credentials] = None
        self._refresh_task: Optional[asyncio.TimerHandle] = None

    async def authenticate(self, force: bool = False) -> Credentials:
        """Authenticate on ecovacs servers."""
        async with self._lock:
            should_login = False
            if self._credentials is None or force:
                _LOGGER.debug("No cached credentials, performing login")
                should_login = True
            elif self._credentials.expires_at < time.time():
                _LOGGER.debug("Credentials have expired, performing login")
                should_login = True

            if should_login:
                self._credentials = await self._auth_client.login()
                if self._refresh_task:
                    self._refresh_task.cancel()

                self._create_refresh_task()

                for on_changed in self._on_credentials_changed:
                    on_changed(self._credentials)

            assert self._credentials is not None
            return self._credentials

    def subscribe(self, callback: Callable[[Credentials], None]) -> None:
        """Add callback on new credentials."""
        self._on_credentials_changed.add(callback)

    def unsubscribe(self, callback: Callable[[Credentials], None]) -> None:
        """Remove callback on new credentials."""
        if callback in self._on_credentials_changed:
            self._on_credentials_changed.remove(callback)

    def _create_refresh_task(self) -> None:
        # refresh at 99% of validity
        assert self._credentials is not None
        validity = (self._credentials.expires_at - time.time()) * 0.99

        self._refresh_task = asyncio.get_event_loop().call_later(
            validity, self._auto_refresh_task
        )

    def _auto_refresh_task(self) -> None:
        _LOGGER.debug("Refresh token")

        async def refresh() -> None:
            try:
                self._refresh_task = None
                await self.authenticate(True)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.error(
                    "An exception occurred during refreshing token", exc_info=True
                )

        asyncio.create_task(refresh())



"""Play sound commands."""

from .common import _ExecuteCommand


class PlaySound(_ExecuteCommand):
    """Play sound command."""

    name = "playSound"

    def __init__(self) -> None:
        super().__init__({"count": 1, "sid": 30})



import pytest

from deebot_client.events import BatteryEvent
from deebot_client.messages import OnBattery
from tests.messages import assert_message


@pytest.mark.parametrize("percentage", [0, 49, 100])
def test_getBattery(percentage: int):
    data = {
        "header": {
            "pri": 1,
            "tzm": 480,
            "ts": "1304637391896",
            "ver": "0.0.1",
            "fwVer": "1.8.2",
            "hwVer": "0.1.1",
        },
        "body": {"data": {"value": percentage, "isLow": 1 if percentage < 20 else 0}},
    }

    assert_message(OnBattery, data, BatteryEvent(percentage))



"""Water info event module."""
from dataclasses import dataclass

from ..util import DisplayNameIntEnum
from .base import Event


class WaterAmount(DisplayNameIntEnum):
    """Enum class for all possible water amounts."""

    LOW = 1
    MEDIUM = 2
    HIGH = 3
    ULTRAHIGH = 4


@dataclass(frozen=True)
class WaterInfoEvent(Event):
    """Water info event representation."""

    mop_attached: bool
    amount: WaterAmount



"""Messages module."""


from ..message import Message
from ..messages.stats import ReportStats
from .battery import OnBattery

# fmt: off
# ordered by file asc
_MESSAGES: list[type[Message]] = [
    OnBattery,

    ReportStats
]
# fmt: on

MESSAGES: dict[str, type[Message]] = {message.name: message for message in _MESSAGES}



"""Stats commands."""
from typing import Any

from ..events import StatsEvent, TotalStatsEvent
from ..message import HandlingResult
from .common import EventBus, _NoArgsCommand


class GetStats(_NoArgsCommand):
    """Get stats command."""

    name = "getStats"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        stats_event = StatsEvent(
            area=data.get("area"),
            time=data.get("time"),
            type=data.get("type"),
        )
        event_bus.notify(stats_event)
        return HandlingResult.success()


class GetTotalStats(_NoArgsCommand):
    """Get stats command."""

    name = "getTotalStats"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        stats_event = TotalStatsEvent(data["area"], data["time"], data["count"])
        event_bus.notify(stats_event)
        return HandlingResult.success()



"""Custom command module."""
from typing import Any, Union

from ..events import CustomCommandEvent
from ..logging_filter import get_logger
from ..message import HandlingState
from .common import CommandResult, EventBus

_LOGGER = get_logger(__name__)


class CustomCommand:
    """Custom command, used when user wants to execute a command, which is not part of this library."""

    def __init__(self, name: str, args: Union[dict, list, None] = None) -> None:
        self._name = name
        if args is None:
            args = {}
        self._args = args

    @property
    def name(self) -> str:
        """Command name."""
        return self._name

    @property
    def args(self) -> Union[dict[str, Any], list]:
        """Command additional arguments."""
        return self._args

    def handle_requested(
        self, events: EventBus, response: dict[str, Any]
    ) -> CommandResult:
        """Handle response from a manual requested command.

        :return: A message response
        """
        if response.get("ret") == "ok":
            data = response.get("resp", response)
            events.notify(CustomCommandEvent(self.name, data))
            return CommandResult.success()

        _LOGGER.warning('Command "%s" was not successfully: %s', self.name, response)
        return CommandResult(HandlingState.FAILED)



"""Position command module."""

from typing import Any

from ..events import Position, PositionsEvent, PositionType
from ..message import HandlingResult
from .common import CommandWithHandling, EventBus


class GetPos(CommandWithHandling):
    """Get volume command."""

    name = "getPos"

    def __init__(self) -> None:
        super().__init__(["chargePos", "deebotPos"])

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        positions = []

        for type_str in ["deebotPos", "chargePos"]:
            data_positions = data.get(type_str, [])

            if isinstance(data_positions, dict):
                positions.append(
                    Position(
                        type=PositionType(type_str),
                        x=data_positions["x"],
                        y=data_positions["y"],
                    )
                )
            else:
                for entry in data_positions:
                    positions.append(
                        Position(
                            type=PositionType(type_str), x=entry["x"], y=entry["y"]
                        )
                    )

        if positions:
            event_bus.notify(PositionsEvent(positions=positions))
            return HandlingResult.success()

        return HandlingResult.analyse()



"""Models module."""
import os
from dataclasses import dataclass
from enum import IntEnum, unique
from typing import Optional, Union

from aiohttp import ClientSession


class DeviceInfo(dict):
    """Class holds all values, which we get from api. Common values can be accessed through properties."""

    @property
    def company(self) -> str:
        """Return company."""
        return str(self["company"])

    @property
    def did(self) -> str:
        """Return did."""
        return str(self["did"])

    @property
    def name(self) -> str:
        """Return name."""
        return str(self["name"])

    @property
    def nick(self) -> Optional[str]:
        """Return nick name."""
        return self.get("nick", None)

    @property
    def resource(self) -> str:
        """Return resource."""
        return str(self["resource"])

    @property
    def device_name(self) -> str:
        """Return device name."""
        return str(self["deviceName"])

    @property
    def status(self) -> int:
        """Return device status."""
        return int(self["status"])

    @property
    def get_class(self) -> str:
        """Return device class."""
        return str(self["class"])


@dataclass(frozen=True)
class Room:
    """Room representation."""

    subtype: str
    id: int
    coordinates: str


@unique
class VacuumState(IntEnum):
    """Vacuum state representation."""

    IDLE = 1
    CLEANING = 2
    RETURNING = 3
    DOCKED = 4
    ERROR = 5
    PAUSED = 6


@dataclass(frozen=True)
class Credentials:
    """Credentials representation."""

    token: str
    user_id: str
    expires_at: int = 0


def _str_to_bool_or_cert(value: Union[bool, str]) -> Union[bool, str]:
    """Convert string to bool or certificate."""
    if isinstance(value, bool):
        return value

    if value is not None:
        value = value.lower()
        if value in ("y", "yes", "t", "true", "on", "1"):
            return True
        if value in ("n", "no", "f", "false", "off", "0"):
            return False
        if os.path.exists(str(value)):
            # User could provide a path to a CA Cert as well, which is useful for Bumper
            if os.path.isfile(str(value)):
                return value
            raise ValueError(f"Certificate path provided is not a file: {value}")

    raise ValueError(f'Cannot convert "{value}" to a bool or certificate path')


class Configuration:
    """Configuration representation."""

    def __init__(
        self,
        session: ClientSession,
        *,
        device_id: str,
        country: str,
        continent: str,
        verify_ssl: Union[bool, str] = True,
    ):
        self._session = session
        self._device_id = device_id
        self._country = country
        self._continent = continent
        self._verify_ssl = _str_to_bool_or_cert(verify_ssl)

    @property
    def session(self) -> ClientSession:
        """Client session."""
        return self._session

    @property
    def device_id(self) -> str:
        """Device id."""
        return self._device_id

    @property
    def country(self) -> str:
        """Country code."""
        return self._country

    @property
    def continent(self) -> str:
        """Continent code."""
        return self._continent

    @property
    def verify_ssl(self) -> Union[bool, str]:
        """Return bool or path to cert."""
        return self._verify_ssl



"""Deebot client exception module."""


class DeebotError(Exception):
    """Deebot error."""


class AuthenticationError(DeebotError):
    """Authentication error."""


class InvalidAuthenticationError(AuthenticationError):
    """Invalid authentication error."""


class ApiError(DeebotError):
    """Api error."""


class MapError(DeebotError):
    """Map error."""


class NotInitializedError(DeebotError):
    """Thrown when not class was not initialized correctly."""



"""Relocate commands."""
from .common import _ExecuteCommand


class SetRelocationState(_ExecuteCommand):
    """Set relocation state command."""

    name = "setRelocationState"

    def __init__(self) -> None:
        super().__init__({"mode": "manu"})



"""Event emitter module."""
import asyncio
import threading
from typing import Awaitable, Callable, Final, Generic, Optional, TypeVar

from ..command import Command
from ..logging_filter import get_logger
from ..models import VacuumState
from . import Event, StatusEvent

_LOGGER = get_logger(__name__)

T = TypeVar("T", bound=Event)


class EventListener(Generic[T]):
    """Object that allows event consumers to easily unsubscribe from event bus."""

    def __init__(
        self,
        event_processing_data: "_EventProcessingData",
        callback: Callable[[T], Awaitable[None]],
    ) -> None:
        self._event_processing_data: Final = event_processing_data
        self.callback: Final = callback

    def unsubscribe(self) -> None:
        """Unsubscribe from event bus."""
        self._event_processing_data.subscribers.remove(self)


class _EventProcessingData(Generic[T]):
    """Data class, which holds all needed data per EventDto."""

    def __init__(self) -> None:
        super().__init__()

        self._subscribers: Final[list[EventListener[T]]] = []
        self._semaphore: Final = asyncio.Semaphore(1)
        self.last_event: Optional[T] = None

    @property
    def subscribers(self) -> list[EventListener[T]]:
        """Return subscribers."""
        return self._subscribers

    @property
    def semaphore(self) -> asyncio.Semaphore:
        """Return semaphore."""
        return self._semaphore


class EventBus:
    """A very simple event bus system."""

    def __init__(self, execute_command: Callable[[Command], Awaitable[None]]):
        self._event_processing_dict: dict[type[Event], _EventProcessingData] = {}
        self._lock = threading.Lock()
        self._execute_command: Final = execute_command

    def has_subscribers(self, event: type[T]) -> bool:
        """Return True, if emitter has subscribers."""
        return (
            len(self._event_processing_dict[event].subscribers) > 0
            if event in self._event_processing_dict
            else False
        )

    def subscribe(
        self,
        event_type: type[T],
        callback: Callable[[T], Awaitable[None]],
    ) -> EventListener[T]:
        """Subscribe to event."""
        event_processing_data = self._get_or_create_event_processing_data(event_type)

        listener = EventListener(event_processing_data, callback)
        event_processing_data.subscribers.append(listener)

        if event_processing_data.last_event:
            # Notify subscriber directly with the last event
            asyncio.create_task(listener.callback(event_processing_data.last_event))
        elif len(event_processing_data.subscribers) == 1:
            # first subscriber therefore do refresh
            self.request_refresh(event_type)

        return listener

    def notify(self, event: T) -> bool:
        """Notify subscriber with given event representation."""
        event_processing_data = self._get_or_create_event_processing_data(type(event))

        if (
            isinstance(event, StatusEvent)
            and event.state == VacuumState.IDLE
            and event_processing_data.last_event
            and event_processing_data.last_event.available == event.available  # type: ignore[attr-defined]
            and event_processing_data.last_event.state == VacuumState.DOCKED  # type: ignore[attr-defined]
        ):
            # todo distinguish better between docked and idle and outside event bus. # pylint: disable=fixme
            # Problem getCleanInfo will return state=idle, when bot is charging
            event = StatusEvent(event.available, VacuumState.DOCKED)  # type: ignore[assignment]

        if event == event_processing_data.last_event:
            _LOGGER.debug("Event is the same! Skipping (%s)", event)
            return False

        event_processing_data.last_event = event
        if event_processing_data.subscribers:
            _LOGGER.debug("Notify subscribers with %s", event)
            for subscriber in event_processing_data.subscribers:
                asyncio.create_task(subscriber.callback(event))
            return True

        _LOGGER.debug("No subscribers... Discharging %s", event)
        return False

    def request_refresh(self, event_class: type[T]) -> None:
        """Request manual refresh."""
        if self.has_subscribers(event_class):
            asyncio.create_task(self._call_refresh_function(event_class))

    async def _call_refresh_function(self, event_class: type[T]) -> None:
        semaphore = self._event_processing_dict[event_class].semaphore
        if semaphore.locked():
            _LOGGER.debug("Already refresh function running. Skipping...")
            return

        async with semaphore:
            from deebot_client.events.const import (  # pylint: disable=import-outside-toplevel
                EVENT_DTO_REFRESH_COMMANDS,
            )

            commands = EVENT_DTO_REFRESH_COMMANDS.get(event_class, [])
            if not commands:
                return

            if len(commands) == 1:
                await self._execute_command(commands[0])
            else:
                tasks = []
                for command in commands:
                    tasks.append(asyncio.create_task(self._execute_command(command)))

                await asyncio.gather(*tasks)

    def _get_or_create_event_processing_data(
        self, event_class: type[T]
    ) -> _EventProcessingData[T]:
        with self._lock:
            event_processing_data = self._event_processing_dict.get(event_class, None)

            if event_processing_data is None:
                event_processing_data = _EventProcessingData()
                self._event_processing_dict[event_class] = event_processing_data

            return event_processing_data



"""Base command."""
from abc import ABC, abstractmethod
from typing import Any, Union


class Command(ABC):
    """Abstract command object."""

    def __init__(self, args: Union[dict, list, None] = None) -> None:
        if args is None:
            args = {}
        self._args = args

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Command name."""
        raise NotImplementedError

    @property
    def args(self) -> Union[dict[str, Any], list]:
        """Command additional arguments."""
        return self._args

    def __eq__(self, obj: object) -> bool:
        if isinstance(obj, Command):
            return self.name == obj.name and self.args == obj.args

        return False

    def __hash__(self) -> int:
        return hash(self.name) + hash(self.args)



"""Charge state commands."""
from typing import Any, Optional

from ..events import StatusEvent
from ..message import HandlingResult
from ..models import VacuumState
from .common import EventBus, _NoArgsCommand
from .const import CODE


class GetChargeState(_NoArgsCommand):
    """Get charge state command."""

    name = "getChargeState"

    @classmethod
    def _handle_body_data_dict(
        cls, event_bus: EventBus, data: dict[str, Any]
    ) -> HandlingResult:
        """Handle message->body->data and notify the correct event subscribers.

        :return: A message response
        """
        if data.get("isCharging") == 1:
            event_bus.notify(StatusEvent(True, VacuumState.DOCKED))
        return HandlingResult.success()

    @classmethod
    def _handle_body(cls, event_bus: EventBus, body: dict[str, Any]) -> HandlingResult:
        if body.get(CODE, 0) == 0:
            # Call this also if code is not in the body
            return super()._handle_body(event_bus, body)

        status: Optional[VacuumState] = None
        if body.get("msg", None) == "fail":
            if body["code"] == "30007":  # Already charging
                status = VacuumState.DOCKED
            elif body["code"] == "5":  # Busy with another command
                status = VacuumState.ERROR
            elif body["code"] == "3":  # Bot in stuck state, example dust bin out
                status = VacuumState.ERROR

        if status:
            event_bus.notify(StatusEvent(True, VacuumState.DOCKED))
            return HandlingResult.success()

        return HandlingResult.analyse()



import pytest

from deebot_client.commands import GetAdvancedMode, SetAdvancedMode
from deebot_client.events import AdvancedModeEvent
from tests.commands import assert_command_requested, assert_set_command
from tests.helpers import get_request_json


@pytest.mark.parametrize("value", [False, True])
def test_get_advanced_mode_requested(value: bool):
    json = get_request_json({"enable": 1 if value else 0})
    assert_command_requested(GetAdvancedMode(), json, AdvancedModeEvent(value))


@pytest.mark.parametrize("value", [False, True])
def test_set_advanced_mode(value: bool):
    args = {"enable": 1 if value else 0}
    assert_set_command(SetAdvancedMode(value), args, AdvancedModeEvent(value))



from typing import Any
from unittest.mock import Mock

from deebot_client.events import Event
from deebot_client.events.event_bus import EventBus
from deebot_client.message import HandlingState, Message


def assert_message(message: type[Message], data: dict[str, Any], expected_event: Event):
    event_bus = Mock(spec_set=EventBus)

    result = message.handle(event_bus, data)

    assert result.state == HandlingState.SUCCESS
    event_bus.notify.assert_called_once_with(expected_event)



"""Setup file."""
from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as file:
    long_description = file.read()

with open("requirements.txt", encoding="utf-8") as file:
    install_requires = list(val.strip() for val in file)

setup(
    name="deebot-client",
    version="1.3.0",
    url="https://github.com/DeebotUniverse/client.py",
    description="a library for controlling certain deebot vacuums",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="DeebotUniverse",
    author_email="deebotuniverse@knatschig-as-hell.info",
    license="GPL-3.0",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Home Automation",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="home automation vacuum robot deebot ecovacs",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    package_data={"deebot_client": ["py.typed"]},
    install_requires=install_requires,
)



"""Util module."""
from __future__ import annotations

import hashlib
from enum import IntEnum, unique
from typing import TYPE_CHECKING, Callable, Iterable, Mapping, TypeVar, overload

from typing_extensions import SupportsIndex

if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem


def md5(text: str) -> str:
    """Hash text using md5."""
    return hashlib.md5(bytes(str(text), "utf8")).hexdigest()


@unique
class DisplayNameIntEnum(IntEnum):
    """Int enum with a property "display_name"."""

    def __new__(cls, *args: int, **_: Mapping) -> DisplayNameIntEnum:
        """Create new DisplayNameIntEnum."""
        obj = int.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, value: int, display_name: str | None = None):
        super().__init__()
        self._value_ = value
        self._display_name = display_name

    @property
    def display_name(self) -> str:
        """Return the custom display name or the lowered name property."""
        if self._display_name:
            return self._display_name

        return self.name.lower()

    @classmethod
    def get(cls, value: str) -> DisplayNameIntEnum:
        """Get enum member from name or display_name."""
        value = str(value).upper()
        if value in cls.__members__:
            return cls[value]

        for member in cls:
            if value == member.display_name.upper():
                return member

        raise ValueError(f"'{value}' is not a valid {cls.__name__} member")

    def __eq__(self, x: object) -> bool:
        if not isinstance(x, type(self)):
            return False
        return bool(self._value_ == x._value_)

    def __ne__(self, x: object) -> bool:
        return not self.__eq__(x)

    def __hash__(self) -> int:
        return hash(self._value_)


_T = TypeVar("_T")


class OnChangedList(list[_T]):
    """List, which will call passed on_change if a change happens."""

    def __init__(self, on_change: Callable[[], None], iterable: Iterable[_T] = ()):
        super().__init__(iterable)
        self._on_change = on_change

    def append(self, __object: _T) -> None:
        """Append object to the end of the list."""

        self._on_change()
        super().append(__object)

    def clear(self) -> None:
        """Remove all items from list."""

        self._on_change()
        super().clear()

    def extend(self, __iterable: Iterable[_T]) -> None:
        """Extend list by appending elements from the iterable."""

        self._on_change()
        super().extend(__iterable)

    def insert(self, __index: SupportsIndex, __object: _T) -> None:
        """Insert object before index."""

        self._on_change()
        super().insert(__index, __object)

    def pop(self, __index=...) -> _T:  # type: ignore
        """
        Remove and return item at index (default last).

        Raises IndexError if list is empty or index is out of range.
        """

        self._on_change()
        return super().pop(__index)

    def remove(self, __value: _T) -> None:
        """
        Remove first occurrence of value.

        Raises ValueError if the value is not present.
        """

        self._on_change()
        super().remove(__value)

    @overload
    def __setitem__(self, i: SupportsIndex, o: _T) -> None:
        ...

    @overload
    def __setitem__(self, s: slice, o: Iterable[_T]) -> None:
        ...

    def __setitem__(self, i, o) -> None:  # type: ignore
        self._on_change()
        super().__setitem__(i, o)

    def __delitem__(self, i: SupportsIndex | slice) -> None:
        self._on_change()
        super().__delitem__(i)

    def __add__(self, x: list[_T]) -> list[_T]:
        self._on_change()
        return super().__add__(x)


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class OnChangedDict(dict[_KT, _VT]):
    """Dict, which will call passed on_change if a change happens."""

    def __init__(
        self, on_change: Callable[[], None], iterable: Iterable[tuple[_KT, _VT]] = ()
    ):
        super().__init__(iterable)
        self._on_change = on_change

    def clear(self) -> None:
        """Remove all items."""

        self._on_change()
        super().clear()

    @overload
    def pop(self, key: _KT) -> _VT:
        ...

    @overload
    def pop(self, key: _KT, default: _VT | _T = ...) -> _VT | _T:
        ...

    def pop(self, key, default=...):  # type: ignore
        """
        Remove specified key and return the corresponding value.

        If the key is not found, return the default if given; otherwise,
        raise a KeyError.
        """

        self._on_change()
        return super().pop(key, default)

    def popitem(self) -> tuple[_KT, _VT]:
        """
        Remove and return a (key, value) pair as a 2-tuple.

        Pairs are returned in LIFO (last-in, first-out) order.
        Raises KeyError if the dict is empty.
        """
        self._on_change()
        return super().popitem()

    @overload
    def update(self, __m: SupportsKeysAndGetItem[_KT, _VT], **kwargs: _VT) -> None:
        ...

    @overload
    def update(self, __m: Iterable[tuple[_KT, _VT]], **kwargs: _VT) -> None:
        ...

    @overload
    def update(self, **kwargs: _VT) -> None:
        ...

    def update(self, __m, **kwargs) -> None:  # type: ignore
        """Update dict."""
        self._on_change()
        super().update(__m, **kwargs)

    def __setitem__(self, k: _KT, v: _VT) -> None:
        self._on_change()
        super().__setitem__(k, v)

    def __delitem__(self, v: _KT) -> None:
        self._on_change()
        super().__delitem__(v)



"""Map module."""
import ast
import asyncio
import base64
import dataclasses
import lzma
import math
import struct
import zlib
from io import BytesIO
from typing import Awaitable, Callable, Final, Optional, Union

from numpy import ndarray, reshape, zeros
from PIL import Image, ImageDraw, ImageOps

from .command import Command
from .commands import GetCachedMapInfo, GetMinorMap
from .events import (
    MajorMapEvent,
    MapSetEvent,
    MapSetType,
    MapSubsetEvent,
    MapTraceEvent,
    MinorMapEvent,
    Position,
    PositionsEvent,
    PositionType,
    RoomsEvent,
)
from .events.event_bus import EventBus, EventListener
from .exceptions import MapError
from .logging_filter import get_logger
from .models import Room
from .util import OnChangedDict, OnChangedList

_LOGGER = get_logger(__name__)
_PIXEL_WIDTH = 50
_POSITION_PNG = {
    PositionType.DEEBOT: "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",  # nopep8
    PositionType.CHARGER: "iVBORw0KGgoAAAANSUhEUgAAAAoAAAAOCAYAAAAWo42rAAAAdUlEQVQoU2NkQAP/nzD8BwkxyjAwIkuhcEASRCmEKYKZhGwq3ER0ReiKSVOIyzRkU8EmwhUyKzAwSNyHyL9QZGD4+wDMBLmVEasimFHIiuEKpcHBhwmeQryBMJFohcjuw2s1SBKHZ8BWo/gauyshvobJEYoZAEOSPXnhzwZnAAAAAElFTkSuQmCC",  # nopep8
}
_OFFSET = 400
_TRACE_MAP = "trace_map"
_COLORS = {
    0x01: "#badaff",  # floor
    0x02: "#4e96e2",  # wall
    0x03: "#1a81ed",  # carpet
    _TRACE_MAP: "#FFFFFF",
    MapSetType.VIRTUAL_WALLS: "#FF0000",
    MapSetType.NO_MOP_ZONES: "#FFA500",
}


def _decompress_7z_base64_data(data: str) -> bytes:
    _LOGGER.debug("[decompress7zBase64Data] Begin")
    final_array = bytearray()

    # Decode Base64
    decoded = base64.b64decode(data)

    i = 0
    for idx in decoded:
        if i == 8:
            final_array += b"\x00\x00\x00\x00"
        final_array.append(idx)
        i += 1

    dec = lzma.LZMADecompressor(lzma.FORMAT_AUTO, None, None)
    decompressed_data = dec.decompress(final_array)

    _LOGGER.debug("[decompress7zBase64Data] Done")
    return decompressed_data


def _calc_value(value: int, min_value: int, max_value: int) -> int:
    try:
        if value is not None:
            new_value = int((int(value) / _PIXEL_WIDTH) + _OFFSET)
            # return value inside min and max
            return min(max_value, max(min_value, new_value))

    except (ZeroDivisionError, ValueError):
        pass

    return min_value or 0


def _calc_point(
    x: int, y: int, image_box: Optional[tuple[int, int, int, int]]
) -> tuple[int, int]:
    if image_box is None:
        image_box = (0, 0, x, y)

    return (
        _calc_value(x, image_box[0], image_box[2]),
        _calc_value(y, image_box[1], image_box[3]),
    )


def _draw_positions(
    positions: list[Position], image: Image, image_box: tuple[int, int, int, int]
) -> None:
    for position in positions:
        icon = Image.open(BytesIO(base64.b64decode(_POSITION_PNG[position.type])))
        image.paste(
            icon,
            _calc_point(position.x, position.y, image_box),
            icon.convert("RGBA"),
        )


def _draw_subset(
    subset: MapSubsetEvent,
    draw: "DashedImageDraw",
    image_box: tuple[int, int, int, int],
) -> None:
    coordinates_ = ast.literal_eval(subset.coordinates)
    points: list[tuple[int, int]] = []
    for i in range(0, len(coordinates_), 2):
        points.append(_calc_point(coordinates_[i], coordinates_[i + 1], image_box))

    if len(points) == 4:
        # close rectangle
        points.append(points[0])

    draw.dashed_line(points, dash=(3, 2), fill=_COLORS[subset.type], width=1)


class Map:
    """Map representation."""

    RESIZE_FACTOR = 3

    def __init__(
        self, execute_command: Callable[[Command], Awaitable[None]], event_bus: EventBus
    ):
        self._execute_command = execute_command
        self._event_bus = event_bus

        self._map_data: Final[MapData] = MapData()
        self._amount_rooms: int = 0
        self._last_image: Optional[LastImage] = None
        self._listeners: list[EventListener] = []

        async def on_map_set(event: MapSetEvent) -> None:
            if event.type == MapSetType.ROOMS:
                self._amount_rooms = len(event.subsets)
                for room_id, _ in self._map_data.rooms.copy().items():
                    if room_id not in event.subsets:
                        self._map_data.rooms.pop(room_id, None)
            else:
                for subset_id, subset in self._map_data.map_subsets.copy().items():
                    if subset.type == event.type and subset_id not in event.subsets:
                        self._map_data.map_subsets.pop(subset_id, None)

        self._event_bus.subscribe(MapSetEvent, on_map_set)

        async def on_map_subset(event: MapSubsetEvent) -> None:
            if event.type == MapSetType.ROOMS and event.subtype:
                room = Room(event.subtype, event.id, event.coordinates)
                if self._map_data.rooms.get(event.id, None) != room:
                    self._map_data.rooms[room.id] = room

                    if len(self._map_data.rooms) == self._amount_rooms:
                        self._event_bus.notify(
                            RoomsEvent(list(self._map_data.rooms.values()))
                        )

            elif self._map_data.map_subsets.get(event.id, None) != event:
                self._map_data.map_subsets[event.id] = event

        self._event_bus.subscribe(MapSubsetEvent, on_map_subset)

    # ---------------------------- METHODS ----------------------------

    def _update_trace_points(self, data: str) -> None:
        _LOGGER.debug("[_update_trace_points] Begin")
        trace_points = _decompress_7z_base64_data(data)

        for i in range(0, len(trace_points), 5):
            byte_position_x = struct.unpack("<h", trace_points[i : i + 2])
            byte_position_y = struct.unpack("<h", trace_points[i + 2 : i + 4])

            # Add To List
            position_x = (int(byte_position_x[0] / 5)) + 400
            position_y = (int(byte_position_y[0] / 5)) + 400

            self._map_data.trace_values.append(position_x)
            self._map_data.trace_values.append(position_y)

        _LOGGER.debug("[_update_trace_points] finish")

    def _draw_map_pieces(self, draw: ImageDraw.Draw) -> None:
        _LOGGER.debug("[_draw_map_pieces] Draw")
        image_x = 0
        image_y = 0

        for i in range(64):
            if i > 0:
                if i % 8 != 0:
                    image_y += 100
                else:
                    image_x += 100
                    image_y = 0

            current_piece = self._map_data.map_pieces[i]
            if current_piece.in_use:
                for x in range(100):
                    current_column = current_piece.points[x]
                    for y in range(100):
                        pixel_type = current_column[y]
                        point_x = image_x + x
                        point_y = image_y + y
                        if (point_x > 6400) or (point_y > 6400):
                            _LOGGER.error(
                                "[get_base64_map] Map Limit 6400!! X: %d Y: %d",
                                point_x,
                                point_y,
                            )
                            raise MapError("Map Limit reached!")
                        if pixel_type in [0x01, 0x02, 0x03]:
                            draw.point((point_x, point_y), fill=_COLORS[pixel_type])

    def enable(self) -> None:
        """Enable map."""
        if self._listeners:
            return

        asyncio.create_task(self._execute_command(GetCachedMapInfo()))

        async def on_position(event: PositionsEvent) -> None:
            self._map_data.positions = event.positions

        self._listeners.append(self._event_bus.subscribe(PositionsEvent, on_position))

        async def on_map_trace(event: MapTraceEvent) -> None:
            if event.start == 0:
                self._map_data.trace_values.clear()

            self._update_trace_points(event.data)

        self._listeners.append(self._event_bus.subscribe(MapTraceEvent, on_map_trace))

        async def on_major_map(event: MajorMapEvent) -> None:
            tasks = []
            for idx, value in enumerate(event.values):
                if (
                    self._map_data.map_pieces[idx].crc32_indicates_update(value)
                    and event.requested
                ):
                    tasks.append(
                        asyncio.create_task(
                            self._execute_command(
                                GetMinorMap(map_id=event.map_id, piece_index=idx)
                            )
                        )
                    )

            if tasks:
                await asyncio.gather(*tasks)

        self._listeners.append(self._event_bus.subscribe(MajorMapEvent, on_major_map))

        async def on_minor_map(event: MinorMapEvent) -> None:
            self._map_data.map_pieces[event.index].update_points(event.value)

        self._listeners.append(self._event_bus.subscribe(MinorMapEvent, on_minor_map))

    def disable(self) -> None:
        """Disable map."""
        listeners = self._listeners
        self._listeners.clear()
        for listener in listeners:
            listener.unsubscribe()

    def refresh(self) -> None:
        """Manually refresh map."""
        if not self._listeners:
            raise MapError("Please enable the map first")

        # todo make it nice pylint: disable=fixme
        self._event_bus.request_refresh(PositionsEvent)
        self._event_bus.request_refresh(MapTraceEvent)
        self._event_bus.request_refresh(MajorMapEvent)

    def get_base64_map(self, width: Optional[int] = None) -> bytes:
        """Return map as base64 image string."""
        if not self._listeners:
            raise MapError("Please enable the map first")

        if (
            self._last_image is not None
            and width == self._last_image.width
            and not self._map_data.changed
        ):
            _LOGGER.debug("[get_base64_map] No need to update")
            return self._last_image.base64_image

        _LOGGER.debug("[get_base64_map] Begin")
        image = Image.new("RGBA", (6400, 6400))
        draw = DashedImageDraw(image)

        self._draw_map_pieces(draw)

        # Draw Trace Route
        if len(self._map_data.trace_values) > 0:
            _LOGGER.debug("[get_base64_map] Draw Trace")
            draw.line(self._map_data.trace_values, fill=_COLORS[_TRACE_MAP], width=1)

        image_box = image.getbbox()
        for subset in self._map_data.map_subsets.values():
            _draw_subset(subset, draw, image_box)

        del draw

        _draw_positions(self._map_data.positions, image, image_box)

        _LOGGER.debug("[get_base64_map] Crop Image")
        cropped = image.crop(image_box)
        del image

        _LOGGER.debug("[get_base64_map] Flipping Image")
        cropped = ImageOps.flip(cropped)

        _LOGGER.debug(
            "[get_base64_map] Map current Size: X: %d Y: %d",
            cropped.size[0],
            cropped.size[1],
        )

        new_size = None
        if width is not None and width > 0:
            height = int((width / cropped.size[0]) * cropped.size[1])
            _LOGGER.debug(
                "[get_base64_map] Resize based on the requested width: %d and calculated height %d",
                width,
                height,
            )
            new_size = (width, height)
        elif cropped.size[0] > 400 or cropped.size[1] > 400:
            _LOGGER.debug(
                "[get_base64_map] Resize disabled.. map over 400 and image width was passed"
            )
        else:
            resize_factor = Map.RESIZE_FACTOR
            _LOGGER.debug("[get_base64_map] Resize factor: %d", resize_factor)
            new_size = (
                cropped.size[0] * resize_factor,
                cropped.size[1] * resize_factor,
            )

        if new_size is not None:
            cropped = cropped.resize(new_size, Image.NEAREST)

        _LOGGER.debug("[get_base64_map] Saving to buffer")
        buffered = BytesIO()
        cropped.save(buffered, format="PNG")
        del cropped

        base64_image = base64.b64encode(buffered.getvalue())
        self._map_data.reset_changed()
        self._last_image = LastImage(base64_image, width)
        _LOGGER.debug("[get_base64_map] Finish")

        return base64_image


class MapPiece:
    """Map piece representation."""

    _NOT_INUSE_CRC32: int = 1295764014

    def __init__(self, on_change: Callable[[], None], index: int) -> None:
        self._on_change = on_change
        self._index = index
        self._points: Optional[ndarray] = None
        self._crc32: int = MapPiece._NOT_INUSE_CRC32

    def crc32_indicates_update(self, crc32: str) -> bool:
        """Return True if update is required."""
        crc32_int = int(crc32)
        if crc32_int == MapPiece._NOT_INUSE_CRC32:
            self._crc32 = crc32_int
            self._points = None
            return False

        return self._crc32 != crc32_int

    @property
    def in_use(self) -> bool:
        """Return True if piece is in use."""
        return self._crc32 != MapPiece._NOT_INUSE_CRC32

    @property
    def points(self) -> ndarray:
        """I'm the 'x' property."""
        if not self.in_use or self._points is None:
            return zeros((100, 100))
        return self._points

    def update_points(self, base64_data: str) -> None:
        """Add map piece points."""
        decoded = _decompress_7z_base64_data(base64_data)
        old_crc32 = self._crc32
        self._crc32 = zlib.crc32(decoded)

        if self._crc32 != old_crc32:
            self._on_change()

        if self.in_use:
            self._points = reshape(list(decoded), (100, 100))
        else:
            self._points = None

    def __hash__(self) -> int:
        """Calculate hash on index and crc32."""
        return hash(self._index) + hash(self._crc32)

    def __eq__(self, obj: object) -> bool:
        if not isinstance(obj, MapPiece):
            return False

        return self._crc32 == obj._crc32 and self._index == obj._index


class DashedImageDraw(ImageDraw.ImageDraw):  # type: ignore
    """Class extend ImageDraw by dashed line."""

    # pylint: disable=invalid-name
    # Copied from https://stackoverflow.com/a/65893631 Credits ands

    def _thick_line(
        self,
        xy: list[tuple[int, int]],
        direction: list[tuple[int, int]],
        fill: Optional[Union[tuple, str]] = None,
        width: int = 0,
    ) -> None:
        if xy[0] != xy[1]:
            self.line(xy, fill=fill, width=width)
        else:
            x1, y1 = xy[0]
            delta_x = direction[1][0] - direction[0][0]
            delta_y = direction[1][1] - direction[0][1]

            if delta_x < 0:
                y1 -= 1

            if delta_y < 0:
                x1 -= 1

            if delta_y != 0:
                if delta_x != 0:
                    k = -delta_x / delta_y
                    a = 1 / math.sqrt(1 + k**2)
                    b = (width * a - 1) / 2
                else:
                    k = 0
                    b = (width - 1) / 2
                x1 = x1 - math.floor(b)
                y1 = y1 - int(k * b)
                x2 = x1 + math.ceil(b)
                y2 = y1 + int(k * b)
            else:
                y1 = y1 - math.floor((width - 1) / 2)
                x2 = x1
                y2 = y1 + math.ceil((width - 1) / 2)
            self.line([(x1, y1), (x2, y2)], fill=fill, width=1)

    def dashed_line(
        self,
        xy: list[tuple[int, int]],
        dash: tuple = (2, 2),
        fill: Optional[Union[tuple, str]] = None,
        width: int = 0,
    ) -> None:
        """Draw a dashed line, or a connected sequence of line segments."""
        # pylint: disable=too-many-locals
        for i in range(len(xy) - 1):
            x1, y1 = xy[i]
            x2, y2 = xy[i + 1]
            x_length = x2 - x1
            y_length = y2 - y1
            length = math.sqrt(x_length**2 + y_length**2)
            dash_enabled = True
            position = 0
            while position <= length:
                for dash_step in dash:
                    if dash_enabled:
                        start = position / length
                        end = min((position + dash_step - 1) / length, 1)
                        self._thick_line(
                            [
                                (
                                    round(x1 + start * x_length),
                                    round(y1 + start * y_length),
                                ),
                                (
                                    round(x1 + end * x_length),
                                    round(y1 + end * y_length),
                                ),
                            ],
                            xy,
                            fill,
                            width,
                        )
                    dash_enabled = not dash_enabled
                    position += dash_step


@dataclasses.dataclass(frozen=True)
class LastImage:
    """Last created image."""

    base64_image: bytes
    width: Optional[int]


class MapData:
    """Map data."""

    def __init__(self) -> None:
        self._changed: bool = False

        def on_change() -> None:
            self._changed = True

        self._on_change = on_change
        self._map_pieces: OnChangedList[MapPiece] = OnChangedList(
            on_change, [MapPiece(on_change, i) for i in range(64)]
        )
        self._map_subsets: OnChangedDict[int, MapSubsetEvent] = OnChangedDict(on_change)
        self._positions: OnChangedList[Position] = OnChangedList(on_change)
        self._rooms: OnChangedDict[int, Room] = OnChangedDict(on_change)
        self._trace_values: OnChangedList[int] = OnChangedList(on_change)

    @property
    def changed(self) -> bool:
        """Indicate if data was changed."""
        return self._changed

    @property
    def map_pieces(self) -> OnChangedList[MapPiece]:
        """Return map pieces."""
        return self._map_pieces

    @property
    def map_subsets(self) -> dict[int, MapSubsetEvent]:
        """Return map subsets."""
        return self._map_subsets

    @property
    def positions(self) -> list[Position]:
        """Return positions."""
        return self._positions

    @positions.setter
    def positions(self, value: list[Position]) -> None:
        if not isinstance(value, OnChangedList):
            value = OnChangedList(self._on_change, value)
        self._positions = value
        self._changed = True

    @property
    def rooms(self) -> dict[int, Room]:
        """Return rooms."""
        return self._rooms

    @property
    def trace_values(self) -> OnChangedList[int]:
        """Return trace values."""
        return self._trace_values

    def reset_changed(self) -> None:
        """Reset changed value."""
        self._changed = False



import pytest

from deebot_client.commands import GetContinuousCleaning, SetContinuousCleaning
from deebot_client.events import ContinuousCleaningEvent
from tests.commands import assert_command_requested, assert_set_command
from tests.helpers import get_request_json


@pytest.mark.parametrize("value", [False, True])
def test_get_continuous_cleaning_requested(value: bool):
    json = get_request_json({"enable": 1 if value else 0})
    assert_command_requested(
        GetContinuousCleaning(), json, ContinuousCleaningEvent(value)
    )


@pytest.mark.parametrize("value", [False, True])
def test_set_continuous_cleaning(value: bool):
    args = {"enable": 1 if value else 0}
    assert_set_command(
        SetContinuousCleaning(value), args, ContinuousCleaningEvent(value)
    )



"""Continuous cleaning (break point) command module."""

from ..events import ContinuousCleaningEvent
from .common import SetEnableCommand, _GetEnableCommand


class GetContinuousCleaning(_GetEnableCommand):
    """Get continuous cleaning command."""

    name = "getBreakPoint"
    event_type = ContinuousCleaningEvent


class SetContinuousCleaning(SetEnableCommand):
    """Set continuous cleaning command."""

    name = "setBreakPoint"
    get_command = GetContinuousCleaning
